##
# File:    CactvsMolecularFactoryTests.py
# Author:  jdw
# Date:    25-Jan-2020
# Version: 0.001
#
# Updates:
#
##
"""
A collection of tests of CactvsMolecularFactory to compare assigned and computed features.

"""
import logging
import os
import unittest

from rcsb.utils.chem.IoUtils import IoUtils
from rcsb.utils.chem.CactvsMoleculeFactory import CactvsMoleculeFactory
from rcsb.utils.chem.MoleculeAnnotationsCompare import MoleculeAnnotationsCompare


HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class CactvsMoleculeFactoryTests(unittest.TestCase):
    skipFlag = True

    def setUp(self):
        #
        self.__workPath = os.path.join(HERE, "test-output")
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cactvsPythonPath = "/apps/lib/cspy"
        self.__molLimit = 50
        self.__exportFlag = True
        #

    def tearDown(self):
        pass

    @unittest.skipIf(skipFlag, "Special install CACTVS dependency")
    def testFromSdf(self):
        aroModel = "daylight"
        retD = {}
        maC = MoleculeAnnotationsCompare()
        ctvsmf = CactvsMoleculeFactory(self.__cactvsPythonPath, self.__workPath)
        ioU = IoUtils()
        sdfL = []
        rdCcObjL = ioU.getComponentDefinitions(os.path.join(self.__dataPath, "components-abbrev.cif"))
        self.assertGreaterEqual(len(rdCcObjL), 4)
        for rdCcObj in rdCcObjL:
            ccId = rdCcObj.getName()
            _, sdfS, atomIdxD = ioU.makeSdf(rdCcObj)
            if self.__exportFlag:
                fp = os.path.join(self.__workPath, rdCcObj.getName() + ".sdf")
                with open(fp, "w", encoding="utf-8") as ofh:
                    ofh.write("%s" % sdfS)
                ctvsmf.setFile(ccId, fp, atomIdxD=atomIdxD)
                cactvsJsonPath = os.path.join(self.__workPath, rdCcObj.getName() + ".json")
                ctvsmf.annotate(cactvsJsonPath, aroModel=aroModel)
                #
                tstFD = ctvsmf.getMoleculeFeatures(aroModel=aroModel)
                refFD = maC.getChemCompFeatures(rdCcObj, descriptorProgram="CACTVS")
                ok, retCmp = maC.compare(refFD, tstFD, tstInfo="cactvs default")
                if not ok:
                    retD[ccId] = retCmp
                #
            sdfL.append(sdfS)
        #
        logger.info("Components processed %d differences %d", len(rdCcObjL), len(retD))
        # logger.info("\n%s", "\n".join(sdfL))


def suiteCactvsBuildTests():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(CactvsMoleculeFactoryTests("testFromSdf"))
    return suiteSelect


if __name__ == "__main__":
    #
    mySuite1 = suiteCactvsBuildTests()
    unittest.TextTestRunner(verbosity=2).run(mySuite1)
    #



##
# File: OeCommonUtils.py
# Date: 22-Oct-2020
#
# Updates:
#
#   1-Mar-2021 jdw  Move aromaticity constrain to relaxed criteria and add additional
#                   contraints to strict criteria (degree and ring membership) and
#                   add an exact criteria which further adds the hcount constraint to
#                   the atom expression
##
import logging

from openeye import oechem

logger = logging.getLogger(__name__)


class OeCommonUtils(object):
    @staticmethod
    def getAtomBondExprOpts(matchOpts):
        """Return the OE constants for atom and bond matching criteria.

        Args:
            matchOpts (string): qualitative description of atom and bond matching criteria

        Returns:
            tuple: OE atom and bond comparison expressions
        """
        if matchOpts in ["exact", "graph-exact", "sub-struct-graph-exact"]:
            atomexpr = (
                oechem.OEExprOpts_AtomicNumber
                | oechem.OEExprOpts_FormalCharge
                | oechem.OEExprOpts_Chiral
                | oechem.OEExprOpts_Aromaticity
                | oechem.OEExprOpts_RingMember
                | oechem.OEExprOpts_Degree
                | oechem.OEExprOpts_HCount
            )
            bondexpr = oechem.OEExprOpts_BondOrder | oechem.OEExprOpts_Chiral | oechem.OEExprOpts_Aromaticity | oechem.OEExprOpts_RingMember

        elif matchOpts in ["default", "graph-default", "strict", "graph-strict", "sub-struct-graph-strict"]:
            atomexpr = (
                oechem.OEExprOpts_AtomicNumber
                | oechem.OEExprOpts_FormalCharge
                | oechem.OEExprOpts_Chiral
                | oechem.OEExprOpts_Aromaticity
                | oechem.OEExprOpts_RingMember
                | oechem.OEExprOpts_Degree
            )
            bondexpr = oechem.OEExprOpts_BondOrder | oechem.OEExprOpts_Chiral | oechem.OEExprOpts_Aromaticity | oechem.OEExprOpts_RingMember
        elif matchOpts in ["relaxed-stereo", "graph-relaxed-stereo", "sub-struct-graph-relaxed-stereo"]:
            atomexpr = oechem.OEExprOpts_AtomicNumber | oechem.OEExprOpts_FormalCharge | oechem.OEExprOpts_Chiral | oechem.OEExprOpts_Aromaticity
            bondexpr = oechem.OEExprOpts_BondOrder | oechem.OEExprOpts_Chiral | oechem.OEExprOpts_Aromaticity
        elif matchOpts in ["relaxed-stereo-sdeq", "graph-relaxed-stereo-sdeq", "sub-struct-graph-relaxed-stereo-sdeq"]:
            atomexpr = oechem.OEExprOpts_AtomicNumber | oechem.OEExprOpts_FormalCharge | oechem.OEExprOpts_Chiral | oechem.OEExprOpts_Aromaticity
            bondexpr = oechem.OEExprOpts_BondOrder | oechem.OEExprOpts_Chiral | oechem.OEExprOpts_EqSingleDouble
        elif matchOpts in ["relaxed", "graph-relaxed", "simple", "sub-struct-graph-relaxed"]:
            atomexpr = oechem.OEExprOpts_AtomicNumber | oechem.OEExprOpts_FormalCharge | oechem.OEExprOpts_Aromaticity
            bondexpr = oechem.OEExprOpts_BondOrder | oechem.OEExprOpts_Aromaticity
        elif matchOpts in ["graph-relaxed-sdeq", "sub-struct-graph-relaxed-sdeq"]:
            atomexpr = oechem.OEExprOpts_AtomicNumber | oechem.OEExprOpts_FormalCharge | oechem.OEExprOpts_Aromaticity
            bondexpr = oechem.OEExprOpts_BondOrder | oechem.OEExprOpts_EqSingleDouble
        else:
            atomexpr = oechem.OEExprOpts_DefaultAtoms
            bondexpr = oechem.OEExprOpts_DefaultBonds
            logger.error("Unanticipated match options %r (applying defaults)", matchOpts)
        #
        #
        # oechem.OEExprOpts_EqSingleDouble
        # atomexpr = oechem.OEExprOpts_AtomicNumber|oechem.OEExprOpts_EqAromatic
        # bondexpr = 0
        #
        # atomexpr = oechem.OEExprOpts_AtomicNumber|oechem.OEExprOpts_Aromaticity
        # bondexpr = oechem.OEExprOpts_BondOrder|oechem.OEExprOpts_EqNotAromatic
        #
        return atomexpr, bondexpr



##
# File:    ChemAxonDescriptorProviderTests.py
# Author:  J. Westbrook
# Date:    17-Aug-2021
# Version: 0.001
#
# Update:
#
#
##
"""
Utilities to deliver ChemAxon rendered chemical descriptors for chemical component definitions.
"""
import logging
import os
import platform
import resource
import time
import unittest

from importlib.metadata import version as get_package_version
from rcsb.utils.chem.ChemAxonDescriptorProvider import ChemAxonDescriptorProvider

__version__ = get_package_version("rcsb.utils.chem")

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class ChemAxonDescriptorProviderTests(unittest.TestCase):
    buildFull = False

    def setUp(self):
        self.__startTime = time.time()
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        #
        logger.debug("Running tests on version %s", __version__)
        logger.info("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testChemAxonDescriptorAbbrev(self):
        """Test construction of abbreviated ChemAxon mapping file."""
        self.__buildMoleculeCacheFiles(buildFlag=True, ccUrlTarget=self.__ccUrlTarget, birdUrlTarget=self.__birdUrlTarget, molLimit=10, useCache=False, ccFileNamePrefix="cc-abbrev")

    def testChemAxonDescriptorCacheFull(self):
        """Test get cached full ChemAxon mapping file."""
        self.__buildMoleculeCacheFiles(buildFlag=False, useCache=True, ccFileNamePrefix="cc-full")

    @unittest.skipUnless(buildFull, "Troubleshooting test")
    def testChemAxonDescriptorFull(self):
        """Test construction of full ChemAxon mapping file."""
        self.__buildMoleculeCacheFiles(buildFlag=True, useCache=False, ccFileNamePrefix="cc-full")

    def __buildMoleculeCacheFiles(self, buildFlag=False, **kwargs):
        """Test build mapping and chemical component cache files from the input component dictionaries"""
        molLimit = kwargs.get("molLimit", None)
        useCache = kwargs.get("useCache", True)
        ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "cc-full")
        ccUrlTarget = kwargs.get("ccUrlTarget", None)
        birdUrlTarget = kwargs.get("birdUrlTarget", None)
        #
        caxP = ChemAxonDescriptorProvider(
            ccUrlTarget=ccUrlTarget, birdUrlTarget=birdUrlTarget, cachePath=self.__cachePath, useCache=useCache, molLimit=molLimit, ccFileNamePrefix=ccFileNamePrefix
        )
        if buildFlag:
            caxP.buildDescriptors()
            ok = caxP.testCache()
            #
            ok = caxP.updateDescriptors(useCache=True)
            self.assertTrue(ok)
        #
        caxP = ChemAxonDescriptorProvider(
            ccUrlTarget=ccUrlTarget, birdUrlTarget=birdUrlTarget, cachePath=self.__cachePath, useCache=True, molLimit=molLimit, ccFileNamePrefix=ccFileNamePrefix
        )
        ok = caxP.testCache()
        self.assertTrue(ok)

        return ok


def buildCacheFiles():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(ChemAxonDescriptorProviderTests("testChemAxonDescriptorAbbrev"))
    suiteSelect.addTest(ChemAxonDescriptorProviderTests("testChemAxonDescriptorCacheFull"))
    suiteSelect.addTest(ChemAxonDescriptorProviderTests("testChemAxonDescriptorFull"))
    return suiteSelect


if __name__ == "__main__":

    mySuite = buildCacheFiles()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    OeMoleculeProviderTests.py
# Author:  J. Westbrook
# Date:    1-Oct-2019
# Version: 0.001
#
# Update:
#
#
##
"""
Tests for utilities to read and process OE molecule instances and related databases created
from of PDB chemical component definitions.

"""
import logging
import os
import time
import unittest

from importlib.metadata import version as get_package_version
from rcsb.utils.chem.OeMoleculeProvider import OeMoleculeProvider

__version__ = get_package_version("rcsb.utils.chem")

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class OeMoleculeProviderTests(unittest.TestCase):
    skipFlag = True

    def setUp(self):
        self.__startTime = time.time()
        #
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        #
        logger.debug("Running tests on version %s", __version__)
        logger.info("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testBuildMoleculeCacheFilesAbbrev(self):
        for molBuildType in ["oe-iso-smiles", "model-xyz", "ideal-xyz"]:
            self.__testBuildMoleculeCacheFiles(
                ccUrlTarget=self.__ccUrlTarget,
                birdUrlTarget=self.__birdUrlTarget,
                ccFileNamePrefix="cc-abbrev",
                oeFileNamePrefix="oe-abbrev",
                molLimit=24,
                fpTypeList=["TREE", "PATH", "MACCS", "CIRCULAR", "LINGO"],
                screenTypeList=["SMARTS"],
                molBuildType=molBuildType,
            )

    @unittest.skipIf(skipFlag, "Long test")
    def testBuildMoleculeCacheFilesSubset(self):
        """Test construction of OE resource files for a subset (1K) of full public chemical reference dictionary"""
        for molBuildType in ["oe-iso-smiles", "model-xyz", "ideal-xyz"]:
            self.__testBuildMoleculeCacheFiles(
                molBuildType=molBuildType,
                molLimit=1000,
                ccFileNamePrefix="cc-1k",
                oeFileNamePrefix="oe-1k",
                screenTypeList=["SMARTS"],
                fpTypeList=["TREE", "PATH", "MACCS", "CIRCULAR", "LINGO"],
            )
        #

    @unittest.skipIf(skipFlag, "Long test")
    def testBuildMoleculeCacheFilesFiltered(self):
        """Test construction of OE resource files for a filtered subset of full public chemical reference dictionary"""
        for molBuildType in ["oe-iso-smiles", "model-xyz", "ideal-xyz"]:
            self.__testBuildMoleculeCacheFiles(
                molBuildType=molBuildType,
                molLimit=1000,
                ccFileNamePrefix="cc-filtered",
                oeFileNamePrefix="oe-filtered",
                screenTypeList=["SMARTS"],
                fpTypeList=["TREE", "PATH", "MACCS", "CIRCULAR", "LINGO"],
            )
        #

    @unittest.skipIf(skipFlag, "Long test")
    def testBuildMoleculeCacheFilesFull(self):
        """Test construction of OE resource files for the full public chemical reference dictionary

        For 31K cc's the build time is ~3123 secs (4 procs)

        """
        for molBuildType in ["oe-iso-smiles", "model-xyz", "ideal-xyz"]:
            self.__testBuildMoleculeCacheFiles(
                molBuildType=molBuildType, ccFileNamePrefix="cc-full", oeFileNamePrefix="oe-full", fpTypeList=["TREE", "PATH", "MACCS", "CIRCULAR", "LINGO"]
            )

    @unittest.skipIf(skipFlag, "Long test")
    def testBuildMoleculeCacheFilesFullScreened(self):
        """Test construction of OE resource files with screened substructure database for the full public chemical reference dictionary"""
        for molBuildType in ["oe-iso-smiles"]:
            self.__testBuildMoleculeCacheFiles(
                molBuildType=molBuildType,
                ccFileNamePrefix="cc-full",
                oeFileNamePrefix="oe-full",
                screenTypeList=["SMARTS"],
                fpTypeList=["TREE", "PATH", "MACCS", "CIRCULAR", "LINGO"],
            )

    def __testBuildMoleculeCacheFiles(self, **kwargs):
        """Test build OE cache files from full component dictionary"""
        ccUrlTarget = kwargs.get("ccUrlTarget", None)
        birdUrlTarget = kwargs.get("birdUrlTarget", None)
        molLimit = kwargs.get("molLimit", 0)
        quietFlag = kwargs.get("quietFlag", True)
        molBuildType = kwargs.get("molBuildType", "ideal-xyz")
        fpTypeList = kwargs.get("fpTypeList", ["TREE"])
        screenTypeList = kwargs.get("screenTypeList", [])
        ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "cc")
        oeFileNamePrefix = kwargs.get("oeFileNamePrefix", "oe")
        #
        startTime = time.time()

        oemp = OeMoleculeProvider(
            ccUrlTarget=ccUrlTarget,
            birdUrlTarget=birdUrlTarget,
            cachePath=self.__cachePath,
            ccFileNamePrefix=ccFileNamePrefix,
            oeFileNamePrefix=oeFileNamePrefix,
            molBuildType=molBuildType,
            useCache=False,
            quietFlag=quietFlag,
            fpTypeList=fpTypeList,
            screenTypeList=screenTypeList,
        )
        ok = oemp.testCache()
        self.assertTrue(ok)
        oemp = OeMoleculeProvider(
            ccUrlTarget=ccUrlTarget,
            birdUrlTarget=birdUrlTarget,
            cachePath=self.__cachePath,
            ccFileNamePrefix=ccFileNamePrefix,
            oeFileNamePrefix=oeFileNamePrefix,
            molBuildType=molBuildType,
            useCache=True,
        )
        endTime = time.time()
        logger.info(">> Completed load molBuildType %r molLimit %r (%.4f seconds)", molBuildType, molLimit, endTime - startTime)
        #
        # ---

        deltaMol = 2
        minMol = minNumFp = molLimit - deltaMol if molLimit else 30000
        for fpType in fpTypeList:
            fpDb = oemp.getFingerPrintDb(fpType="TREE")
            logger.debug("fpType %r length %d", fpType, fpDb.NumFingerPrints())
            self.assertGreaterEqual(fpDb.NumFingerPrints(), minNumFp)
        #
        ccId = "004"
        oeMol = oemp.getMol(ccId)
        logger.debug("%s atom count %d", ccId, len(list(oeMol.GetAtoms())))
        #
        if molBuildType in ["oe-iso-smiles"]:
            self.assertGreaterEqual(len(list(oeMol.GetAtoms())), 12)
        else:
            self.assertGreaterEqual(len(list(oeMol.GetAtoms())), 20)
        #
        oeDb, oeDbIdx = oemp.getOeMolDatabase()
        logger.debug("Type db %r length %d type idx %r length %d", type(oeDb), oeDb.NumMols(), type(oeDbIdx), len(oeDbIdx))
        self.assertGreaterEqual(oeDb.NumMols(), minMol)
        self.assertGreaterEqual(len(oeDbIdx), minMol)
        #
        if molBuildType in ["oe-iso-smiles"] and screenTypeList:
            ssDb = oemp.getSubSearchDb()
            self.assertGreaterEqual(ssDb.NumMolecules(), minMol)
        return True


def buildCacheFiles():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(OeMoleculeProviderTests("testBuildMoleculeCacheFilesAbbrev"))
    suiteSelect.addTest(OeMoleculeProviderTests("testBuildMoleculeCacheFilesSubset"))
    suiteSelect.addTest(OeMoleculeProviderTests("testBuildMoleculeCacheFilesFiltered"))
    suiteSelect.addTest(OeMoleculeProviderTests("testBuildMoleculeCacheFilesFull"))
    return suiteSelect


if __name__ == "__main__":

    mySuite = buildCacheFiles()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
#
# File:    OeDepictAllTests.py
# Author:  jdw
# Date:    5-May-2013
# Version: 0.001
#
# Updates:
#  4-May-2014 jdw add example for depiction from SMILES input
#  6-Jun-2016 jdw general cleanup
##
"""
A collection of tests for the OEDepict and related classes.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Creative Commons Attribution 3.0 Unported"

import logging
import os
import os.path
import time
import unittest

from rcsb.utils.chem.OeDepict import OeDepict
from rcsb.utils.chem.OeMoleculeProvider import OeMoleculeProvider

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class OeDepictAllTests(unittest.TestCase):
    def setUp(self):
        self.__startTime = time.time()
        self.__workPath = os.path.join(HERE, "test-output")
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        #
        self.__oeMolD = self.__getCache()

    def tearDown(self):
        pass

    def __getCache(self):
        oemp = OeMoleculeProvider(
            ccUrlTarget=self.__ccUrlTarget,
            birdUrlTarget=self.__birdUrlTarget,
            ccFileNamePrefix="cc-abbrev",
            cachePath=self.__cachePath,
            molBuildType="model-xyz",
            useCache=True,
            oeFileNamePrefix="oe-abbrev",
        )
        ok = oemp.testCache()
        self.assertTrue(ok)
        return oemp.getOeMolD()

    def testDepictAll(self):
        """Test case -  single OE molecule depiction."""
        try:
            for ccId, oeMol in self.__oeMolD.items():
                imagePath = os.path.join(self.__workPath, "image", ccId[0], ccId + ".svg")
                oed = OeDepict()
                title = ""
                oed.setMolTitleList([(ccId, oeMol, title)])
                oed.setDisplayOptions(
                    labelAtomName=False, labelAtomCIPStereo=True, labelAtomIndex=False, labelBondIndex=False, labelBondCIPStereo=True, cellBorders=False, bondDisplayWidth=0.5
                )
                oed.setGridOptions(rows=1, cols=1)
                oed.prepare()
                oed.write(imagePath)
            for ccId, oeMol in self.__oeMolD.items():
                imagePath = os.path.join(self.__workPath, "image_labeled", ccId[0], ccId + ".svg")
                oed = OeDepict()
                title = ""
                oed.setMolTitleList([(ccId, oeMol, title)])
                oed.setDisplayOptions(
                    labelAtomName=True, labelAtomCIPStereo=True, labelAtomIndex=False, labelBondIndex=False, abelBondCIPStereo=True, cellBorders=False, bondDisplayWidth=0.5
                )
                oed.setGridOptions(rows=1, cols=1)
                oed.prepare()
                oed.write(imagePath)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def suiteDepict():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(OeDepictAllTests("testDepictAll"))
    return suiteSelect


if __name__ == "__main__":
    mySuite = suiteDepict()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    OeDepictCompareTests.py
# Author:  jdw
# Date:    28-Oct-2019
# Version: 0.001
#
# Updates:
#
##
"""
A collection of tests to compare assigned and computed features.

"""
import logging
import os
import time
import unittest
from collections import defaultdict

from openeye import oechem
from rcsb.utils.chem.ChemCompIndexProvider import ChemCompIndexProvider
from rcsb.utils.chem.ChemCompMoleculeProvider import ChemCompMoleculeProvider
from rcsb.utils.chem.OeDepict import OeDepict
from rcsb.utils.chem.OeDepictAlign import OeDepictMCSAlignPage
from rcsb.utils.chem.OeMoleculeFactory import OeMoleculeFactory

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class OeDepictCompareTests(unittest.TestCase):
    skipFlag = True

    def setUp(self):
        #
        self.__startTime = time.time()
        self.__workPath = os.path.join(HERE, "test-output")
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        self.__molLimit = 50
        self.__ccIdList = ["002", "004", "PRD_000921"]
        self.__myKwargs = {
            "ccUrlTarget": self.__ccUrlTarget,
            "birdUrlTarget": self.__birdUrlTarget,
            "cachePath": self.__cachePath,
            "useCache": True,
            "ccFileNamePrefix": "cc-abbrev",
            "molLimit": None,
        }
        #

    def tearDown(self):
        pass

    def __getChemCompDefs(self):
        ccMolD = {}
        ccIdxD = {}
        try:
            ccmP = ChemCompMoleculeProvider(**self.__myKwargs)
            ok = ccmP.testCache()
            ccMolD = ccmP.getMolD()
            ccmP = ChemCompIndexProvider(**self.__myKwargs)
            ccIdxD = ccmP.getIndex()
            ok = ccmP.testCache(minCount=10)
            self.assertTrue(ok)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ccMolD, ccIdxD

    def testCompareByBuildType(self):
        """Compare depictions constructed molecules with various builds from chemical defintions -
        all build types 8769 (all)
        connect - smiles 6743
        model vs iso smiles 5937
        ideal va iso smiles  7047
        """
        doDepict = False
        ccResultD = {}
        genResultD = {}
        smilesByBuildTypeD = {}
        try:
            ccMolD, ccIdxD = self.__getChemCompDefs()
            #
            limitPerceptions = True
            # molBuildTypeL = ["model-xyz", "ideal-xyz", "connection-table", "oe-iso-smiles"]
            molBuildTypeL = ["ideal-xyz", "oe-iso-smiles"]
            #
            startTime = time.time()
            oefm = OeMoleculeFactory()
            oefm.setQuiet()
            for molBuildType in molBuildTypeL:
                for ccId, idxD in ccIdxD.items():
                    ccObj = ccMolD[ccId]
                    # ----
                    ccIsoSmiles = idxD["oe-iso-smiles"]
                    ccSmiles = idxD["oe-smiles"]
                    # ----
                    tId = oefm.setChemCompDef(ccObj)
                    if not tId:
                        logger.info("Skipping bad component %r", ccId)
                        continue
                    self.assertEqual(tId, ccId)
                    ok = oefm.build(molBuildType=molBuildType, limitPerceptions=limitPerceptions)
                    if not ok:
                        logger.info("Build using %r failed for %s", molBuildType, ccId)
                        continue
                    # ------
                    oeMol = oefm.getGraphMol()
                    oeIsoSmiles = oefm.getIsoSMILES()
                    oeSmiles = oefm.getCanSMILES()
                    ccEq = oeIsoSmiles == ccIsoSmiles and oeSmiles == ccSmiles
                    #
                    oefmR = OeMoleculeFactory()
                    oefmR.setQuiet()
                    ccIdGen = ccId + "_gen"
                    oefmR.setDescriptor(oeIsoSmiles, "oe-iso-smiles", ccIdGen)
                    ok = oefmR.build(molBuildType="oe-iso-smiles", limitPerceptions=limitPerceptions)
                    if not ok:
                        logger.info("Build using %r failed for %s", molBuildType, ccIdGen)
                        continue
                    # ------
                    #
                    # oeMolGen = oefmR.getGraphMol()
                    oeIsoSmilesGen = oefmR.getIsoSMILES()
                    oeSmilesGen = oefmR.getCanSMILES()
                    genEq = oeIsoSmiles == oeIsoSmilesGen and oeSmiles == oeSmilesGen
                    smilesByBuildTypeD.setdefault(ccId, {}).setdefault(molBuildType, []).append(oeIsoSmilesGen)
                    #
                    logger.debug("%s buildType %s ccEq %r genEq %r", ccId, molBuildType, ccEq, genEq)
                    if not ccEq:
                        ccResultD.setdefault(molBuildType, []).append(ccId)
                    if not genEq:
                        genResultD.setdefault(molBuildType, []).append(ccId)

                    if doDepict:
                        pS = "-limited" if limitPerceptions else ""
                        imagePath = os.path.join(self.__workPath, ccId + "-%s%s.svg" % (molBuildType, pS))
                        oed = OeDepict()
                        title = ""
                        oed.setMolTitleList([(ccId, oeMol, title)])
                        oed.setDisplayOptions(labelAtomName=False, labelAtomCIPStereo=True, labelAtomIndex=False, labelBondIndex=False, cellBorders=False, bondDisplayWidth=0.5)
                        oed.setGridOptions(rows=1, cols=1)
                        oed.prepare()
                        oed.write(imagePath)
            logger.info("Completed comparing %d molecules in %d builds (%.4f seconds)", len(ccIdxD), len(molBuildTypeL), time.time() - startTime)
            #
            #
            for molBuildType in molBuildTypeL:
                if molBuildType in genResultD:
                    logger.info("GEN %s (%d) %r", molBuildType, len(genResultD[molBuildType]), genResultD[molBuildType])

            numDiff = 0
            for ccId, btD in smilesByBuildTypeD.items():
                tS = set()
                for molBuildType, sL in btD.items():
                    tS.add(sL[0])
                if len(tS) > 1:
                    numDiff += 1
                    logger.debug("%s diff smiles (%d) %r", ccId, len(tS), tS)
            logger.info("Components with inconsistent SMILES %d", numDiff)
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()

    @unittest.skipIf(skipFlag, "Troubleshooting test")
    def testCompareDescriptors(self):
        molLimit = self.__molLimit
        # for molBuildType in ["model-xyz", "ideal-xyz", None]:
        for molBuildType in ["model-xyz"]:
            logger.info("Rebuild cache using molBuildType %r (molLimit=%r)", molBuildType, molLimit)
            self.__testReproduceDescriptors(molBuildType, limitPerceptions=True)

    def __displayAlignedPair(self, ccIdRef, oeMolRef, ccIdFit, oeMolFit, title=None):
        oed = OeDepictMCSAlignPage()
        oed.setDisplayOptions(
            labelAtomName=True, labelAtomCIPStereo=True, labelAtomIndex=False, labelBondIndex=False, highlightStyleFit="ballAndStickInverse", bondDisplayWidth=0.5
        )
        oed.setRefMol(oeMolRef, ccIdRef)
        oed.setFitMol(oeMolFit, ccIdFit)
        myTitle = title if title else "compare"
        imgPath = os.path.join(self.__workPath, myTitle + "-" + ccIdRef + "-" + ccIdFit + ".svg")
        # logger.info("Using image path %r", imgPath)
        aML = oed.alignPair(imagePath=imgPath)
        if aML:
            for (rCC, rAt, tCC, tAt) in aML:
                logger.debug("%5s %-5s %5s %-5s", rCC, rAt, tCC, tAt)
        oed = OeDepictMCSAlignPage()
        oed.setDisplayOptions(
            labelAtomName=True, labelAtomCIPStereo=True, labelAtomIndex=False, labelBondIndex=False, highlightStyleFit="ballAndStickInverse", bondDisplayWidth=0.5
        )

    def __testReproduceDescriptors(self, molBuildType, limitPerceptions=True):
        #
        ccMolD, ccIdxD = self.__getChemCompDefs()
        oemf = OeMoleculeFactory()
        countD = defaultdict(int)
        for ccId, ccDef in ccMolD.items():
            tId = oemf.setChemCompDef(ccDef)
            if ccId != tId:
                continue
            oemf.build(molBuildType=molBuildType, limitPerceptions=limitPerceptions)
            oeMol = oemf.getMol()
            #
            countD["total components"] += 1
            if ccId not in ccIdxD:
                logger.info("Missing ccIndex entry for %s", ccId)
                continue
            ccdD = ccIdxD[ccId]
            if ccdD["ambiguous"]:
                countD["ambiguous component"] += 1
                continue
            #
            countD["total molecules"] += 1

            nativeCanIsoSmiles = oechem.OECreateIsoSmiString(oeMol)
            canIsoSmiles = oechem.OEMolToSmiles(oeMol)
            isoSmiles = oemf.getIsoSMILES()
            canSmiles = oemf.getCanSMILES()
            # check interal consistency
            if nativeCanIsoSmiles != isoSmiles:
                logger.error("%s stored and calculated OE smiles differ %s %s", ccId, nativeCanIsoSmiles, isoSmiles)
            if canIsoSmiles != isoSmiles:
                logger.error("%s calculated OE ISO and canonical smiles differ %s %s", ccId, isoSmiles, canIsoSmiles)

            # compare with archived values
            if isoSmiles != ccdD["oe-iso-smiles"]:
                logger.info("%s ISO SMILES differ \nccd: %r  \nOE:  %r", ccId, ccdD["oe-iso-smiles"], isoSmiles)
                countD["iso_smiles_diff"] += 1
            # ----------
            if canSmiles != ccdD["oe-smiles"]:
                logger.info("%s CAN SMILES differ \nccd: %r  \nOE:  %r", ccId, ccdD["oe-smiles"], canSmiles)
                countD["smiles_diff"] += 1

            formula = oemf.getFormula()
            if formula.upper() != ccdD["formula"].upper():
                logger.debug("%s formulas differ \nccd: %r  \nOE:  %r", ccId, ccdD["formula"], formula)
                countD["formula_diff"] += 1
            # ---------
            inchiKey = oemf.getInChIKey()
            if inchiKey != ccdD["inchikey"]:
                logger.debug("%s InChI keys differ \nccd: %r  \nOE:  %r", ccId, ccdD["inchikey"], inchiKey)
                countD["inchikey_diff"] += 1
            #
            inchi = oemf.getInChI()
            if inchi != ccdD["inchi"]:
                logger.debug("%s InChIs differ \nccd: %r  \nOE:  %r", ccId, ccdD["inchi"], inchi)
                countD["inchi_diff"] += 1
        #
        #
        for ky, vl in countD.items():
            logger.info("%-12s %6d", ky, vl)


def suiteCompareDescriptorsTests():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(OeDepictCompareTests("testCompareDescriptors"))
    return suiteSelect


if __name__ == "__main__":
    #
    mySuite1 = suiteCompareDescriptorsTests()
    unittest.TextTestRunner(verbosity=2).run(mySuite1)
    #



##
# File:    OeSubStructSearchUtilsTests.py
# Author:  J. Westbrook
# Date:    2-Oct-2020
# Version: 0.001
#
# Update:
#
#
##
"""
Tests for substructure search modes on core and search index PDB chemical component definitions.

"""
import logging
import os
import time
import unittest

from importlib.metadata import version as get_package_version
from rcsb.utils.chem.ChemCompIndexProvider import ChemCompIndexProvider
from rcsb.utils.chem.ChemCompSearchIndexProvider import ChemCompSearchIndexProvider
from rcsb.utils.chem.OeMoleculeProvider import OeMoleculeProvider
from rcsb.utils.chem.OeSearchMoleculeProvider import OeSearchMoleculeProvider
from rcsb.utils.chem.OeSubStructSearchUtils import OeSubStructSearchUtils

__version__ = get_package_version("rcsb.utils.chem")

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()
logger.setLevel(logging.INFO)


class OeSubStructSearchUtilsTests(unittest.TestCase):
    useFull = False

    def setUp(self):

        self.__workPath = os.path.join(HERE, "test-output")
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        self.__numProcPrep = 8
        self.__numProcSearch = 8
        self.__minCount = None
        self.__startTime = time.time()
        #
        if OeSubStructSearchUtilsTests.useFull:
            self.__myKwargs = {
                "cachePath": self.__cachePath,
                "useCache": True,
                "ccFileNamePrefix": "cc-full",
                "oeFileNamePrefix": "oe-full",
                "molBuildType": "model-xyz",
                "limitPerceptions": False,
                "screenTypeList": None,
                "numProc": self.__numProcPrep,
                "suppressHydrogens": True,
                "matchOpts": "sub-struct-graph-relaxed",
            }
        else:
            self.__myKwargs = {
                "ccUrlTarget": self.__ccUrlTarget,
                "birdUrlTarget": self.__birdUrlTarget,
                "cachePath": self.__cachePath,
                "useCache": True,
                "ccFileNamePrefix": "cc-abbrev",
                "oeFileNamePrefix": "oe-abbrev",
                "molBuildType": "model-xyz",
                "limitPerceptions": False,
                "screenTypeList": None,
                "numProc": self.__numProcPrep,
                "suppressHydrogens": True,
                "matchOpts": "sub-struct-graph-relaxed",
            }
        #
        logger.debug("Running tests on version %s", __version__)
        logger.info("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def __resultContains(self, ccId, matchResultList):
        for matchResult in matchResultList:
            if matchResult.ccId == ccId:
                return True
        return False

    @unittest.skipIf(not useFull, "Requires full data set")
    def testSubStructureSearchFromIndexSelected(self):
        matchOpts = self.__myKwargs.get("matchOpts", "sub-struct-graph-relaxed")
        numProc = self.__numProcSearch
        oemp = OeSearchMoleculeProvider(**self.__myKwargs)
        ok = oemp.testCache()
        self.assertTrue(ok)
        oesU = OeSubStructSearchUtils(oemp)
        #
        ccIdxP = ChemCompSearchIndexProvider(**self.__myKwargs)
        ok = ccIdxP.testCache(minCount=self.__minCount)
        self.assertTrue(ok)
        ccIdxD = ccIdxP.getIndex()
        ky = next(iter(ccIdxD))
        oeMol = oemp.getMol(ky)
        #
        for ccId in ["BNZ", "ALA"]:
            # ----
            startTime = time.time()
            oeMol = oemp.getMol(ccId)
            #
            ccIdL = oesU.prefilterIndex(oeMol, ccIdxP, matchOpts=matchOpts)
            logger.info("%s search length %d in (%.4f seconds)", ccId, len(ccIdL), time.time() - startTime)
            #
            retStatus, mL = oesU.searchSubStructure(oeMol, ccIdList=ccIdL, matchOpts=matchOpts, numProc=numProc)
            logger.info("%s status %r result length %d in (%.4f seconds)", ccId, retStatus, len(mL), time.time() - startTime)
            self.assertTrue(retStatus)
            self.assertTrue(self.__resultContains(ccId, mL))
            # ----

    def testSubStructureSearchFromIndexBase(self):
        matchOpts = self.__myKwargs.get("matchOpts", "sub-struct-graph-relaxed")
        numProc = self.__numProcSearch
        oemp = OeSearchMoleculeProvider(**self.__myKwargs)
        ok = oemp.testCache()
        self.assertTrue(ok)
        oesU = OeSubStructSearchUtils(oemp)
        #
        ccIdxP = ChemCompSearchIndexProvider(**self.__myKwargs)
        ok = ccIdxP.testCache(minCount=self.__minCount)
        self.assertTrue(ok)
        ccIdxD = ccIdxP.getIndex()
        ky = next(iter(ccIdxD))
        oeMol = oemp.getMol(ky)

        numMols = 10
        for ccId, _ in list(ccIdxD.items())[:numMols]:
            # ----
            startTime = time.time()
            oeMol = oemp.getMol(ccId)
            #
            ccIdL = oesU.prefilterIndex(oeMol, ccIdxP, matchOpts=matchOpts)
            logger.info("%s search length %d in (%.4f seconds)", ccId, len(ccIdL), time.time() - startTime)
            #
            retStatus, mL = oesU.searchSubStructure(oeMol, ccIdList=ccIdL, matchOpts=matchOpts, numProc=numProc)
            logger.info("%s status %r result length %d in (%.4f seconds)", ccId, retStatus, len(mL), time.time() - startTime)
            self.assertTrue(retStatus)
            self.assertTrue(self.__resultContains(ccId, mL))
            # ----

    @unittest.skipIf(not useFull, "Requires full data set")
    def testSubStructureSearchBaseSelected(self):
        matchOpts = self.__myKwargs.get("matchOpts", "sub-struct-graph-relaxed")
        numProc = self.__numProcSearch
        oemp = OeMoleculeProvider(**self.__myKwargs)
        ok = oemp.testCache()
        self.assertTrue(ok)
        oesU = OeSubStructSearchUtils(oemp)
        #
        #
        ccIdxP = ChemCompIndexProvider(**self.__myKwargs)
        ok = ccIdxP.testCache(minCount=self.__minCount)
        self.assertTrue(ok)
        ccIdxD = ccIdxP.getIndex()
        #
        ky = next(iter(ccIdxD))
        oeMol = oemp.getMol(ky)
        #
        for ccId in ["BNZ", "ALA"]:
            # ----
            startTime = time.time()
            oeMol = oemp.getMol(ccId)
            #
            ccIdL = oesU.prefilterIndex(oeMol, ccIdxP, matchOpts=matchOpts)

            logger.info("%s search length %d in (%.4f seconds)", ccId, len(ccIdL), time.time() - startTime)
            #
            retStatus, mL = oesU.searchSubStructure(oeMol, ccIdList=ccIdL, matchOpts=matchOpts, numProc=numProc)
            logger.info("%s result length %d in (%.4f seconds)", ccId, len(mL), time.time() - startTime)
            self.assertTrue(retStatus)
            self.assertTrue(self.__resultContains(ccId, mL))
            # ----

    def testSubStructureSearchBase(self):

        matchOpts = self.__myKwargs.get("matchOpts", "sub-struct-graph-relaxed")
        numProc = self.__numProcSearch
        oemp = OeMoleculeProvider(**self.__myKwargs)
        ok = oemp.testCache()
        self.assertTrue(ok)
        oesU = OeSubStructSearchUtils(oemp)
        #
        ccIdxP = ChemCompIndexProvider(**self.__myKwargs)
        ok = ccIdxP.testCache(minCount=self.__minCount)
        self.assertTrue(ok)
        ccIdxD = ccIdxP.getIndex()
        #
        ky = next(iter(ccIdxD))
        oeMol = oemp.getMol(ky)
        #
        numMols = 10
        for ccId, _ in list(ccIdxD.items())[:numMols]:
            # ----
            startTime = time.time()
            oeMol = oemp.getMol(ccId)
            ccIdL = oesU.prefilterIndex(oeMol, ccIdxP, matchOpts=matchOpts)
            logger.info("%s search length %d in (%.4f seconds)", ccId, len(ccIdL), time.time() - startTime)
            #
            retStatus, mL = oesU.searchSubStructure(oeMol, ccIdList=ccIdL, matchOpts=matchOpts, numProc=numProc)
            logger.info("%s result length %d in (%.4f seconds)", ccId, len(mL), time.time() - startTime)
            self.assertTrue(retStatus)
            self.assertTrue(self.__resultContains(ccId, mL))
            # ----

    #  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------


def exhaustiveSubStructureSearch():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(OeSubStructSearchUtilsTests("testSubStructureSearchFromIndex"))
    suiteSelect.addTest(OeSubStructSearchUtilsTests("testSubStructureSearchBase"))
    suiteSelect.addTest(OeSubStructSearchUtilsTests("testSubStructureSearchFromIndexBase"))
    suiteSelect.addTest(OeSubStructSearchUtilsTests("testSubStructureSearchFromIndexSelected"))
    return suiteSelect


if __name__ == "__main__":
    mySuite = exhaustiveSubStructureSearch()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    ObMolecularFactoryTests.py
# Author:  jdw
# Date:    11-Jan-2020
# Version: 0.001
#
# Updates:
#
##
"""
A collection of tests of ObMolecularFactory to compare assigned and computed features.

"""
import logging
import os
import unittest

from rcsb.utils.chem.IoUtils import IoUtils

try:
    from rcsb.utils.chem.ObMoleculeFactory import ObMoleculeFactory
except Exception:
    pass


HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class ObMolecularFactoryTests(unittest.TestCase):
    skipFlag = True

    def setUp(self):
        #
        self.__workPath = os.path.join(HERE, "test-output")
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__molLimit = 50
        self.__exportFlag = True
        #

    def tearDown(self):
        pass

    @unittest.skipIf(skipFlag, "Special install OpenBabel dependency")
    def testFromSdfFile(self):
        obmf = ObMoleculeFactory()
        ioU = IoUtils()
        sdfL = []
        rdCcObjL = ioU.getComponentDefinitions(os.path.join(self.__dataPath, "components-abbrev.cif"))
        self.assertGreaterEqual(len(rdCcObjL), 4)
        for rdCcObj in rdCcObjL:
            ccId = rdCcObj.getName()
            _, sdfS, atomIdxD = ioU.makeSdf(rdCcObj)
            if self.__exportFlag:
                fp = os.path.join(self.__workPath, rdCcObj.getName() + ".sdf")
                with open(fp, "w", encoding="utf-8") as ofh:
                    ofh.write("%s" % sdfS)
                obmf.setFile(ccId, fp, molFormat="mol", atomIdxD=atomIdxD)
                # obmf.updateProperties()
                # obmf.updateStereoChemAssignments()
                # obmf.updateAromaticAssignments()
                # _ = obmf.getRdMoleculeFeatures()
                #
            sdfL.append(sdfS)
        #
        logger.debug("\n%s", "\n".join(sdfL))

    @unittest.skipIf(skipFlag, "Special install OpenBabel dependency")
    def testFromSdfString(self):
        obmf = ObMoleculeFactory()
        ioU = IoUtils()
        sdfL = []
        rdCcObjL = ioU.getComponentDefinitions(os.path.join(self.__dataPath, "components-abbrev.cif"))
        self.assertGreaterEqual(len(rdCcObjL), 4)
        for rdCcObj in rdCcObjL:
            ccId = rdCcObj.getName()
            _, sdfS, atomIdxD = ioU.makeSdf(rdCcObj)
            obmf.setString(ccId, sdfS, molFormat="mol", atomIdxD=atomIdxD)
            obD = obmf.getMoleculeFeatures()
            logger.info("dictionary %r", obD)
            #
            sdfL.append(sdfS)
        #
        logger.debug("\n%s", "\n".join(sdfL))


def suiteObBuildTests():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(ObMolecularFactoryTests("testFromSdfFile"))
    return suiteSelect


if __name__ == "__main__":
    #
    mySuite1 = suiteObBuildTests()
    unittest.TextTestRunner(verbosity=2).run(mySuite1)
    #



##
# File:    BatchChemSearchTests.py
# Author:  J. Westbrook
# Date:    3-Mar-2021
# Version: 0.001
#
# Update:
#
#
##
"""
Tests for batch search mode.

"""
import logging
import os
import platform
import resource
import time
import unittest

from importlib.metadata import version as get_package_version
from rcsb.utils.chem.BatchChemSearch import BatchChemSearch

__version__ = get_package_version("rcsb.utils.chem")


HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()
logger.setLevel(logging.INFO)


class BatchChemSearchTests(unittest.TestCase):
    def setUp(self):
        self.__startTime = time.time()
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        self.__smilesCod = os.path.join(self.__dataPath, "allcod.smi")
        #
        self.__queryResultsPath = os.path.join(self.__cachePath, "batch-descriptor-results.json")
        self.__testFlagFull = False
        #
        os.environ["CHEM_SEARCH_CACHE_PATH"] = os.path.join(self.__cachePath)
        os.environ["CHEM_SEARCH_CC_PREFIX"] = "cc-full" if self.__testFlagFull else "cc-abbrev"
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif") if not self.__testFlagFull else None
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif") if not self.__testFlagFull else None
        #
        self.__numMolsTest = 4000
        self.__numProc = 4
        logger.debug("Running tests on version %s", __version__)
        logger.info("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testBatchChemSearchSetup(self):
        """Test case - COD SMILES search """
        try:
            bsw = BatchChemSearch(ccUrlTarget=self.__ccUrlTarget, birdUrlTarget=self.__birdUrlTarget, cachePath=self.__cachePath, numProc=self.__numProc)
            ok = bsw.testCache()
            self.assertTrue(ok)
            smiL = bsw.fetchDescriptorList(self.__smilesCod, swap=True)
            logger.info("Query length (%d)", len(smiL))
            #
            smiL = bsw.splitSmiles(smiL)
            retL = bsw.doQuery(smiL[: self.__numMolsTest], "SMILES", matchOpts="graph-exact")
            logger.info("Result length (%d)", len(retL))
            self.assertGreater(len(retL), 2)
            for ii, ret in enumerate(retL, 1):
                logger.debug("%5d %8s %4s (%.3f) %s: %s", ii, ret.queryId, ret.ccId, ret.fpScore, ret.queryType, ret.query)
            #
            ok = bsw.storeMatchList(self.__queryResultsPath, retL)
            self.assertTrue(ok)
            mL = bsw.fetchMatchList(self.__queryResultsPath)
            self.assertEqual(len(mL), len(retL))

        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def batchChemSearchTests():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(BatchChemSearchTests("testBatchChemSearchSetup"))
    return suiteSelect


if __name__ == "__main__":
    mySuite = batchChemSearchTests()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
#
# File:    OeDepictFromChemTests.py
# Author:  jdw
# Date:    5-Mar-2020
# Version: 0.001
#
# Updates:
##
"""
A collection of tests for the OEDepict and related classes working directly from
chemical component definitions

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Creative Commons Attribution 3.0 Unported"

import logging
import os
import os.path
import time
import unittest

from rcsb.utils.chem.ChemCompMoleculeProvider import ChemCompMoleculeProvider
from rcsb.utils.chem.OeDepict import OeDepict
from rcsb.utils.chem.OeMoleculeFactory import OeMoleculeFactory

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class OeDepictFromChemCompTests(unittest.TestCase):
    def setUp(self):
        #
        self.__startTime = time.time()
        self.__workPath = os.path.join(HERE, "test-output")
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        self.__molLimit = 50
        self.__ccIdList = ["002", "004"]
        #

    def tearDown(self):
        pass

    def __getChemCompDefs(self, molLimit=24):
        ccMolD = {}
        try:
            useCache = True
            ccFileNamePrefix = "cc-abbrev"
            ccmP = ChemCompMoleculeProvider(
                ccUrlTarget=self.__ccUrlTarget,
                birdUrlTarget=self.__birdUrlTarget,
                cachePath=self.__cachePath,
                useCache=useCache,
                ccFileNamePrefix=ccFileNamePrefix,
                molLimit=molLimit,
            )
            ok = ccmP.testCache(minCount=molLimit)
            self.assertTrue(ok)
            ccMolD = ccmP.getMolD()
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ccMolD

    def testDepictByBuildType(self):
        """Compare depictions constructed molecules with various builds from chemical defintions -"""
        try:
            ccIdList = self.__ccIdList
            ccMolD = self.__getChemCompDefs()
            #
            limitPerceptions = True
            molBuildTypeL = ["model-xyz", "ideal-xyz", "connection-table", "oe-iso-smiles"]
            #
            startTime = time.time()
            oefm = OeMoleculeFactory()
            for molBuildType in molBuildTypeL:
                for ccId in ccIdList:
                    ccObj = ccMolD[ccId]
                    # ----
                    tId = oefm.setChemCompDef(ccObj)
                    self.assertEqual(tId, ccId)
                    ok = oefm.build(molBuildType=molBuildType, limitPerceptions=limitPerceptions)
                    if not ok:
                        logger.info("Build using %r failed for %s", molBuildType, ccId)
                        continue
                    #
                    oeMol = oefm.getGraphMol()
                    pS = "-limited" if limitPerceptions else ""
                    imagePath = os.path.join(self.__workPath, ccId + "-%s%s.svg" % (molBuildType, pS))
                    oed = OeDepict()
                    title = ""
                    oed.setMolTitleList([(ccId, oeMol, title)])
                    oed.setDisplayOptions(labelAtomName=False, labelAtomCIPStereo=True, labelAtomIndex=False, labelBondIndex=False, cellBorders=False, bondDisplayWidth=0.5)
                    oed.setGridOptions(rows=1, cols=1)
                    oed.prepare()
                    oed.write(imagePath)
            logger.info("Completed depictions on %d molecules (%.4f seconds)", len(ccIdList) * len(molBuildTypeL), time.time() - startTime)
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def suiteDepictFromChemComp():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(OeDepictFromChemCompTests("testDepictByBuildType"))
    return suiteSelect


if __name__ == "__main__":
    mySuite = suiteDepictFromChemComp()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    OeSearchMoleculeProviderTests.py
# Author:  J. Westbrook
# Date:    1-Oct-2019
# Version: 0.001
#
# Update:
#
#
##
"""
Tests for utilities to read and process search OE molecule instances and related databases created
from chemical component search indices.

"""
import logging
import os
import platform
import resource
import time
import unittest


from importlib.metadata import version as get_package_version
from rcsb.utils.chem.OeSearchMoleculeProvider import OeSearchMoleculeProvider

__version__ = get_package_version("rcsb.utils.chem")

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class OeSearchMoleculeProviderTests(unittest.TestCase):
    skipFlag = True

    def setUp(self):
        self.__startTime = time.time()
        #
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        #
        logger.debug("Running tests on version %s", __version__)
        logger.info("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testSearchBuildMoleculeCacheFilesAbbrev(self):
        self.__testBuildSearchMoleculeCacheFiles(
            ccUrlTarget=self.__ccUrlTarget,
            birdUrlTarget=self.__birdUrlTarget,
            ccFileNamePrefix="cc-abbrev",
            oeFileNamePrefix="oe-abbrev",
            molLimit=None,
            fpTypeList=["TREE", "PATH", "MACCS", "CIRCULAR", "LINGO"],
            screenTypeList=["SMARTS"],
            numProc=2,
            minCount=24,
        )

    @unittest.skipIf(skipFlag, "Long troubleshooting test")
    def testSearchBuildMoleculeCacheFilesFull(self):
        self.__testBuildSearchMoleculeCacheFiles(
            ccFileNamePrefix="cc-full",
            oeFileNamePrefix="oe-full",
            molLimit=None,
            minCount=500,
            fpTypeList=["TREE", "PATH", "MACCS", "CIRCULAR", "LINGO"],
            screenTypeList=[],
            numProc=2,
        )

    def __testBuildSearchMoleculeCacheFiles(self, **kwargs):
        """Test build OE cache files from full component dictionary"""
        ccUrlTarget = kwargs.get("ccUrlTarget", None)
        birdUrlTarget = kwargs.get("birdUrlTarget", None)
        molLimit = kwargs.get("molLimit", None)
        quietFlag = kwargs.get("quietFlag", True)
        fpTypeList = kwargs.get("fpTypeList", ["TREE"])
        screenTypeList = kwargs.get("screenTypeList", [])
        ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "cc")
        oeFileNamePrefix = kwargs.get("oeFileNamePrefix", "oe")
        numProc = kwargs.get("numProc", 2)
        minCount = kwargs.get("minCount", 0)
        #
        startTime = time.time()

        # Using abbreviated reference source files
        oesmp = OeSearchMoleculeProvider(
            ccUrlTarget=ccUrlTarget,
            birdUrlTarget=birdUrlTarget,
            cachePath=self.__cachePath,
            ccFileNamePrefix=ccFileNamePrefix,
            oeFileNamePrefix=oeFileNamePrefix,
            useCache=False,
            quietFlag=quietFlag,
            fpTypeList=fpTypeList,
            screenTypeList=screenTypeList,
            numProc=numProc,
            molLimit=molLimit,
        )

        ok = oesmp.testCache()
        self.assertTrue(ok)
        #
        endTime = time.time()
        logger.info(">> Completed load molLimit %r (%.4f seconds)", molLimit, endTime - startTime)
        #
        # ---
        oesmp = OeSearchMoleculeProvider(cachePath=self.__cachePath, ccFileNamePrefix=ccFileNamePrefix, oeFileNamePrefix=oeFileNamePrefix, useCache=True)
        #
        deltaMol = 2
        minMol = minNumFp = minCount - deltaMol if minCount else 10
        for fpType in fpTypeList:
            fpDb = oesmp.getFingerPrintDb(fpType="TREE")
            logger.debug("fpType %r length %d", fpType, fpDb.NumFingerPrints())
            self.assertGreaterEqual(fpDb.NumFingerPrints(), minNumFp)
        #
        ccId = "004"
        oeMol = oesmp.getMol(ccId)
        logger.debug("%s atom count %d", ccId, len(list(oeMol.GetAtoms())))
        #
        self.assertGreaterEqual(len(list(oeMol.GetAtoms())), 12)
        #
        oeDb, oeDbIdx = oesmp.getOeMolDatabase()
        logger.debug("Type db %r length %d type idx %r length %d", type(oeDb), oeDb.NumMols(), type(oeDbIdx), len(oeDbIdx))
        self.assertGreaterEqual(oeDb.NumMols(), minMol)
        self.assertGreaterEqual(len(oeDbIdx), minMol)
        #
        if screenTypeList:
            ssDb = oesmp.getSubSearchDb()
            self.assertGreaterEqual(ssDb.NumMolecules(), minMol)


def buildCacheFiles():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(OeSearchMoleculeProviderTests("testBuildSearchMoleculeCacheFilesAbbrev"))

    return suiteSelect


if __name__ == "__main__":

    mySuite = buildCacheFiles()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:  OeDepictAlign.py
# Date:  27-Oct-2019  J. Westbrook
#
# Updates:
#
##
"""
Classes to depict aligned 2D chemical diagrams in a variety of media formats.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

# pylint: disable=too-many-lines

import logging
import os
import os.path

from openeye import oechem
from openeye import oedepict
from wrapt_timeout_decorator import timeout

from rcsb.utils.chem.OeCommonUtils import OeCommonUtils
from rcsb.utils.chem.OeDepict import OeDepictBase
from rcsb.utils.io.FileUtil import FileUtil

logger = logging.getLogger(__name__)


class OeDepictAlignBase(OeDepictBase):
    """Base class for aligned 2D molecular renderings containing molecular
    object data and common display preferences.
    """

    def __init__(self, **kwargs):
        super(OeDepictAlignBase, self).__init__()
        #
        #
        self.timeOut = kwargs.get("timeOut", None)
        self._refId = None
        self._refPath = None
        self._refMol = None
        self._refTitle = None
        self._refImagePath = None
        #
        self._fitId = None
        self._fitPath = None
        self._fitMol = None
        self._fitTitle = None
        self._fitImagePath = None
        #
        self._pairMolList = []
        #
        self._searchType = "default"
        self._minAtomMatchFraction = 0.50
        self._mcss = None
        self._miter = None
        #
        self._ss = None

    def setRefMol(self, oeMol, ccId, title=None, imagePath=None):
        try:
            self._refId = ccId
            self._refMol = oeMol
            #
            if title is not None:
                self._refMol.SetTitle(title)
                self._refTitle = title
            else:
                self._refMol.SetTitle(self._refId)
                self._refTitle = None
            #
            self._refImagePath = imagePath if imagePath is not None else self._refId + ".svg"
            return True
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))

        return False

    def setFitMol(self, oeMol, ccId, title=None, imagePath=None):
        try:
            self._fitId = ccId
            self._fitMol = oeMol
            #
            if title is not None:
                self._fitMol.SetTitle(title)
                self._fitTitle = title
            else:
                self._fitMol.SetTitle(self._fitId)
                self._fitTitle = None
            #
            self._fitImagePath = imagePath if imagePath is not None else self._fitId + ".svg"
            return True
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return False

    def addFitMolList(self, oeMolList, suppressHydrogens=False, imageDirPath=".", imageFilePrefix=None):
        """Set the list of molecules to be compared with reference molecule by MCSS.

        From the input list build the internal pair list (self._pairMolList) of
        tuples  [(refId,refMol,refTitle,refImagePath,fitId,fitMol,fitTitle,fitImagePath),(),...]
        """
        self._pairMolList = []
        try:
            imageFilePrefix = imageFilePrefix + "-" if imageFilePrefix else ""
            for oeMol in oeMolList:
                refId = self._refId
                refImagePath = self._refImagePath
                refMol = oechem.OESuppressHydrogens(oechem.OEGraphMol(self._refMol)) if suppressHydrogens else self._refMol
                #
                fitId = str(oeMol.GetTitle()).upper()
                fitMol = oechem.OESuppressHydrogens(oechem.OEGraphMol(oeMol)) if suppressHydrogens else oeMol
                fitTitle = fitId + "/" + refId
                refTitle = refId + "/" + fitId
                fitImagePath = os.path.join(imageDirPath, imageFilePrefix + fitId + "-with-ref-" + refId + ".svg")
                self._pairMolList.append((refId, refMol, refTitle, refImagePath, fitId, fitMol, fitTitle, fitImagePath))
            return True
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))

        return False

    def setPairMolList(self, pairOeMolList, suppressHydrogens=False, imageDirPath=".", imageFilePrefix=None):
        """Set the pairs of molecules to be compared by MCSS.

        From the input list build the internal pair list (self._pairMolList) of
        tuples  [(refId,refMol,refTitle,refImagePath,fitId,fitMol,fitTitle,fitImagePath),(),...]
        """
        self._pairMolList = []
        try:
            imageFilePrefix = imageFilePrefix + "-" if imageFilePrefix else ""
            for refId, refOeMol, fitId, fitOeMol in pairOeMolList:
                refImagePath = None
                refMol = oechem.OESuppressHydrogens(oechem.OEGraphMol(refOeMol)) if suppressHydrogens else refOeMol
                #
                fitMol = oechem.OESuppressHydrogens(oechem.OEGraphMol(fitOeMol)) if suppressHydrogens else fitOeMol
                fitTitle = fitId + "/" + refId
                refTitle = refId + "/" + fitId
                fitImagePath = os.path.join(imageDirPath, imageFilePrefix + "pair-fit-" + fitId + "-with-ref-" + refId + ".svg")
                refImagePath = os.path.join(imageDirPath, imageFilePrefix + "pair-fit-" + refId + "-with-ref-" + fitId + ".svg")
                self._pairMolList.append((refId, refMol, refTitle, refImagePath, fitId, fitMol, fitTitle, fitImagePath))
            return True
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))

        return False

    def setSearchType(self, sType="default", minAtomMatchFraction=0.50):
        self._searchType = sType
        self._minAtomMatchFraction = minAtomMatchFraction
        return self._searchType

    def _setupMCSS(self, refmol):
        """Internal initialization for the MCSS comparison."""
        # self._mcss = oechem.OEMCSSearch(oechem.OEMCSType_Approximate)
        self._mcss = oechem.OEMCSSearch(oechem.OEMCSType_Exhaustive)
        #
        atomexpr, bondexpr = OeCommonUtils.getAtomBondExprOpts(self._searchType)

        self._mcss.Init(refmol, atomexpr, bondexpr)
        #
        # self._mcss.SetMCSFunc(OEMCSMaxBondsCompleteCycles())
        # self._mcss.SetMCSFunc(OEMCSMaxAtoms())
        #
        # Half of the reference molecule --
        #
        # nAtomsRef=refmol.NumAtoms()
        # self._mcss.SetMinAtoms(nAtomsRef/2)

    def _setupSubStructure(self, refmol):
        """Internal initialization for a substructure comparison."""

        #
        atomexpr, bondexpr = OeCommonUtils.getAtomBondExprOpts(self._searchType)
        #
        self._ss = oechem.OESubSearch(refmol, atomexpr, bondexpr)
        #

    def _setHighlightStyleRef(self, matchObj, refMol):
        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
        #
        optInverseFit = False
        if self._params["highlightStyleRef"] == "ballAndStick":
            hstyle = oedepict.OEHighlightStyle_BallAndStick
        elif self._params["highlightStyleRef"] == "stick":
            hstyle = oedepict.OEHighlightStyle_Stick
        elif self._params["highlightStyleRef"] == "stickInverse":
            hstyle = oedepict.OEHighlightStyle_Stick
            optInverseFit = True
        elif self._params["highlightStyleRef"] == "ballAndStickInverse":
            hstyle = oedepict.OEHighlightStyle_BallAndStick
            optInverseFit = True
        else:
            hstyle = oedepict.OEHighlightStyle_BallAndStick

        if self._params["highLightMatchColorRef"] == "blue":
            myHighLightMatchColor = oechem.OEBlueTint
        elif self._params["highLightMatchColorRef"] == "green":
            myHighLightMatchColor = oechem.OEGreenTint
        elif self._params["highLightMatchColorRef"] == "pink":
            myHighLightMatchColor = oechem.OEPinkTint
        else:
            myHighLightMatchColor = oechem.OEBlueTint

        if self._params["highLightNotMatchColorRef"] == "blue":
            myHighLightNotMatchColor = oechem.OEBlueTint
        elif self._params["highLightNotMatchColorRef"] == "green":
            myHighLightNotMatchColor = oechem.OEGreenTint
        elif self._params["highLightNotMatchColorRef"] == "pink":
            myHighLightNotMatchColor = oechem.OEPinkTint
        else:
            myHighLightNotMatchColor = oechem.OEBlueTint

        #
        matchedatoms = oechem.OEIsAtomMember(matchObj.GetPatternAtoms())
        matchedbonds = oechem.OEIsBondMember(matchObj.GetPatternBonds())
        if optInverseFit:
            oedepict.OEAddHighlighting(refdisp, myHighLightNotMatchColor, hstyle, oechem.OENotAtom(matchedatoms), oechem.OENotBond(matchedbonds))
        else:
            oedepict.OEAddHighlighting(refdisp, myHighLightMatchColor, hstyle, matchedatoms, matchedbonds)

        return refdisp

    def _setHighlightStyleFit(self, matchObj, fitMol):
        fitdisp = oedepict.OE2DMolDisplay(fitMol, self._opts)

        optInverseFit = False
        if self._params["highlightStyleFit"] == "ballAndStick":
            hstyle = oedepict.OEHighlightStyle_BallAndStick
        elif self._params["highlightStyleFit"] == "stick":
            hstyle = oedepict.OEHighlightStyle_Stick
        elif self._params["highlightStyleFit"] == "stickInverse":
            hstyle = oedepict.OEHighlightStyle_Stick
            optInverseFit = True
        elif self._params["highlightStyleFit"] == "ballAndStickInverse":
            hstyle = oedepict.OEHighlightStyle_BallAndStick
            optInverseFit = True
        else:
            hstyle = oedepict.OEHighlightStyle_BallAndStick

        if self._params["highLightMatchColorFit"] == "blue":
            myHighLightMatchColor = oechem.OEBlueTint
        elif self._params["highLightMatchColorFit"] == "green":
            myHighLightMatchColor = oechem.OEGreenTint
        elif self._params["highLightMatchColorFit"] == "pink":
            myHighLightMatchColor = oechem.OEPinkTint
        else:
            myHighLightMatchColor = oechem.OEBlueTint

        if self._params["highLightNotMatchColorFit"] == "blue":
            myHighLightNotMatchColor = oechem.OEBlueTint
        elif self._params["highLightNotMatchColorFit"] == "green":
            myHighLightNotMatchColor = oechem.OEGreenTint
        elif self._params["highLightNotMatchColorFit"] == "pink":
            myHighLightNotMatchColor = oechem.OEPinkTint
        else:
            myHighLightNotMatchColor = oechem.OEBlueTint

        #
        # matchedatoms = oechem.OEIsAtomMember(matchObj.GetPatternAtoms())
        # matchedbonds = oechem.OEIsBondMember(matchObj.GetPatternBonds())
        matchedatoms = oechem.OEIsAtomMember(matchObj.GetTargetAtoms())
        matchedbonds = oechem.OEIsBondMember(matchObj.GetTargetBonds())

        if optInverseFit:
            oedepict.OEAddHighlighting(fitdisp, myHighLightNotMatchColor, hstyle, oechem.OENotAtom(matchedatoms), oechem.OENotBond(matchedbonds))
        else:
            oedepict.OEAddHighlighting(fitdisp, myHighLightMatchColor, hstyle, matchedatoms, matchedbonds)

        return fitdisp


class OeDepictMCSAlignMultiPage(OeDepictAlignBase):
    """Create 2D depictions of MCSS alignments. Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.

    Output images are rendered in a grid layout that can span multiple pages.
    """

    def __init__(self, **kwargs):
        super(OeDepictMCSAlignMultiPage, self).__init__(**kwargs)
        self.__grid = None
        self.__gridRows = None
        self.__gridCols = None
        self.__multi = None
        self.__image = None
        self.__citer = None
        #

    def alignPairListMulti(self, imagePath="multi.pdf"):
        aM = []
        self._params["gridCols"] = 2
        try:
            aM = self.__alignListMultiWorker(imagePath=imagePath, layout="pairs")
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))

        return aM

    def alignOneWithListMulti(self, imagePath="multi.pdf"):
        aM = []
        try:
            aM = self.__alignListMultiWorker(imagePath=imagePath, layout="list")
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))

        return aM

    def __setupImageMulti(self):
        """Internal method to configure a multipage image."""
        #
        self.__gridRows = self._params["gridRows"]
        self.__gridCols = self._params["gridCols"]
        #
        if self._params["pageOrientation"] == "landscape":
            self.__multi = oedepict.OEMultiPageImageFile(oedepict.OEPageOrientation_Landscape, oedepict.OEPageSize_US_Letter)
        else:
            self.__multi = oedepict.OEMultiPageImageFile(oedepict.OEPageOrientation_Portrait, oedepict.OEPageSize_US_Letter)

        self.__newPage()

    def __newPage(self):
        """Internal method to advance to a new page in a multipage configuration."""
        rows = self.__gridRows
        cols = self.__gridCols
        self.__image = self.__multi.NewPage()
        self.__grid = oedepict.OEImageGrid(self.__image, rows, cols)
        self.__grid.SetCellGap(self._params["cellGap"])
        self.__grid.SetMargins(self._params["cellMargin"])
        logger.debug("Num columns %d", self.__grid.NumCols())
        logger.debug("Num rows    %d", self.__grid.NumRows())
        self._opts = oedepict.OE2DMolDisplayOptions(self.__grid.GetCellWidth(), self.__grid.GetCellHeight(), oedepict.OEScale_AutoScale)
        self._assignDisplayOptions()
        self.__citer = self.__grid.GetCells()

    @timeout("instance.timeOut", use_signals=False, dec_allow_eval=True)
    def __alignListMultiWorker(self, imagePath="multi.pdf", layout="pairs"):
        """Working method comparing a reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,refImagePath,fitId,fitMol,fitTitle,fitImagePath)

        Map of corresponding atoms is returned.

        Image Output is in multipage layout.
        """
        #
        self.__setupImageMulti()
        #
        atomMap = []
        firstOne = True
        iCount = 0
        for (refId, refMol, _, _, fitId, fitMol, fitTitle, _) in self._pairMolList:
            iCount += 1
            #
            oedepict.OEPrepareDepiction(refMol)
            self._setupMCSS(refMol)
            #
            #
            fitMol.SetTitle(fitTitle)
            #
            oedepict.OEPrepareDepiction(fitMol)
            #
            nAtomsRef = refMol.NumAtoms()
            nAtomsFit = fitMol.NumAtoms()
            minAtoms = min(nAtomsRef, nAtomsFit)
            mcssMinAtoms = int(minAtoms * self._minAtomMatchFraction)
            self._mcss.SetMinAtoms(mcssMinAtoms)

            # scaling
            refscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            fitscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            self._opts.SetScale(min(refscale, fitscale))

            unique = True
            self._miter = self._mcss.Match(fitMol, unique)
            logger.debug("mcss match completed for refId %s fitId %s", refId, fitId)
            if self._miter.IsValid():
                match = self._miter.Target()
                oedepict.OEPrepareAlignedDepiction(fitMol, self._mcss.GetPattern(), match)

                # Depict reference molecule with MCS highlighting
                if (firstOne and layout in ["list"]) or layout in ["pairs"]:
                    firstOne = False
                    if layout in ["pairs"]:
                        refdisp = self._setHighlightStyleRef(matchObj=match, refMol=self._mcss.GetPattern())
                    else:
                        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
                    if not self.__citer.IsValid():
                        self.__newPage()
                    cell = self.__citer.Target()
                    self.__citer.Next()
                    oedepict.OERenderMolecule(cell, refdisp)
                    if self._params["cellBorders"]:
                        oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                # Depict fit molecule with MCS highlighting
                fitdisp = self._setHighlightStyleFit(matchObj=match, fitMol=fitMol)
                if not self.__citer.IsValid():
                    self.__newPage()
                cell = self.__citer.Target()
                self.__citer.Next()
                oedepict.OERenderMolecule(cell, fitdisp)

                if self._params["cellBorders"]:
                    oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

                logger.debug("mcss match completed for refId %s fitId %s total map length %d", refId, fitId, len(atomMap))

        logger.debug("writing image %s", imagePath)
        fU = FileUtil()
        fU.mkdirForFile(imagePath)
        oedepict.OEWriteMultiPageImage(imagePath, self.__multi)
        logger.debug("completed with map lenth %d", len(atomMap))
        #
        return atomMap


class OeDepictMCSAlignPage(OeDepictAlignBase):
    """Create 2D depictions of MCSS alignments. Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.

    Output images are rendered to a single page image with grid layout.
    """

    def __init__(self, **kwargs):
        super(OeDepictMCSAlignPage, self).__init__(**kwargs)
        self.__grid = None
        # self.__gridRows = None
        # self.__gridCols = None
        # self.__multi = None
        self.__image = None
        self.__citer = None
        #

    def __setupImage(self):
        """Internal method to configure a single pair alignment image."""
        #
        self.__image = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        rows = self._params["gridRows"]
        cols = self._params["gridCols"]
        self.__grid = oedepict.OEImageGrid(self.__image, rows, cols)
        self.__grid.SetCellGap(self._params["cellGap"])
        self.__grid.SetMargins(self._params["cellMargin"])
        logger.debug("Num columns %d", self.__grid.NumCols())
        logger.debug("Num rows    %d", self.__grid.NumRows())
        self._opts = oedepict.OE2DMolDisplayOptions(self.__grid.GetCellWidth(), self.__grid.GetCellHeight(), oedepict.OEScale_AutoScale)
        self._assignDisplayOptions()
        self.__citer = self.__grid.GetCells()

    def alignPair(self, imagePath="single-pair.png"):
        """Depict a single aligned ref/fit molecule pair or the first ref/fit molecule pair on the
        current _pairMolList.  Display options set for a single grid row with two columns.
        """
        self._params["gridCols"] = 2
        self._params["gridRows"] = 1
        self.__setupImage()
        #
        aM = []
        self._pairMolList = []
        self._pairMolList.append((self._refId, self._refMol, self._refTitle, None, self._fitId, self._fitMol, self._fitTitle, None))
        try:
            aM = self.__alignListWorker(imagePath=imagePath, layout="pairs")
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignPairList(self, imagePath="single-pair-list.png"):
        aM = []
        self._params["gridCols"] = 2
        try:
            aM = self.__alignListWorker(imagePath=imagePath, layout="pairs")
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignOneWithList(self, imagePath="single-list.png"):
        aM = []
        try:
            aM = self.__alignListWorker(imagePath=imagePath, layout="list")
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return aM

    @timeout("instance.timeOut", use_signals=False, dec_allow_eval=True)
    def __alignListWorker(self, imagePath="single.pdf", layout="pairs"):
        """Working method comparing a reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,refImgPath,fitId,fitMol,fitTitle,fitImgPath)

        Map of corresponding atoms is returned.

        Output image is a single-page with grid layout.
        """
        #
        self.__setupImage()
        #
        atomMap = []

        firstOne = True

        for (refId, refMol, _, _, fitId, fitMol, fitTitle, _) in self._pairMolList:
            #
            oedepict.OEPrepareDepiction(refMol)
            self._setupMCSS(refMol)
            #
            title = fitTitle if fitTitle else fitId
            fitMol.SetTitle(title)

            oedepict.OEPrepareDepiction(fitMol)
            #
            #
            nAtomsRef = refMol.NumAtoms()
            nAtomsFit = fitMol.NumAtoms()
            minAtoms = min(nAtomsRef, nAtomsFit)
            self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))

            # scaling
            refscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            fitscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            self._opts.SetScale(min(refscale, fitscale))

            unique = True
            self._miter = self._mcss.Match(fitMol, unique)

            if self._miter.IsValid():
                match = self._miter.Target()
                oedepict.OEPrepareAlignedDepiction(fitMol, self._mcss.GetPattern(), match)

                # Depict reference molecule with MCS highlighting
                if (firstOne and layout in ["list"]) or layout in ["pairs"]:
                    firstOne = False
                    if layout in ["pairs"]:
                        refdisp = self._setHighlightStyleRef(matchObj=match, refMol=self._mcss.GetPattern())
                    else:
                        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
                    if not self.__citer.IsValid():
                        break
                    cell = self.__citer.Target()
                    self.__citer.Next()
                    oedepict.OERenderMolecule(cell, refdisp)
                    if self._params["cellBorders"]:
                        oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                # Depict fit molecule with MCS highlighting
                fitdisp = self._setHighlightStyleFit(matchObj=match, fitMol=fitMol)
                if not self.__citer.IsValid():
                    break
                cell = self.__citer.Target()
                self.__citer.Next()
                oedepict.OERenderMolecule(cell, fitdisp)

                if self._params["cellBorders"]:
                    oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        fU = FileUtil()
        fU.mkdirForFile(imagePath)
        oedepict.OEWriteImage(imagePath, self.__image)
        return atomMap


class OeDepictMCSAlign(OeDepictAlignBase):
    """Create 2D depictions of MCSS alignments. Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.

    Outputs are separate image files with a single diagram per file.
    """

    def __init__(self, **kwargs):
        super(OeDepictMCSAlign, self).__init__(**kwargs)
        #
        self.__imageRef = None
        self.__imageFit = None
        #

    def __setupImage(self):
        """Internal method to configure a single page image."""
        #
        self.__imageRef = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        self.__imageFit = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        self._opts = oedepict.OE2DMolDisplayOptions(self.__imageRef.GetWidth(), self.__imageRef.GetHeight(), oedepict.OEScale_AutoScale)
        self._assignDisplayOptions()

    def alignPair(self):
        """Depict a single aligned ref/fit molecule pair or the first ref/fit molecule pair on the
        current _pairMolList.  Display options set for a single grid row with two columns.
        """
        aM = []
        self._pairMolList = []
        self._pairMolList.append((self._refId, self._refMol, self._refTitle, self._refImagePath, self._fitId, self._fitMol, self._fitTitle, self._fitImagePath))
        try:
            aM = self.__alignListWorker(layout="pairs")
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignPairList(self):
        aM = []
        try:
            aM = self.__alignListWorker(layout="pairs")
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignOneWithList(self):
        aM = []
        try:
            aM = self.__alignListWorker(layout="list")
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return aM

    @timeout("instance.timeOut", use_signals=False, dec_allow_eval=True)
    def __alignListWorker(self, layout="pairs"):
        """Working method comparing a reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,refImagePath, fitId,fitMol,fitTitle, fitImagePath)

        Map of corresponding atoms is returned.

        Writes separate output images for the reference and fit molecules for each comparison pair.
        """
        #
        atomMap = []
        firstOne = True
        fU = FileUtil()

        for (refId, refMol, _, refImagePath, fitId, fitMol, fitTitle, fitImagePath) in self._pairMolList:
            #
            self.__setupImage()
            #
            oedepict.OEPrepareDepiction(refMol)
            #
            self._setupMCSS(refMol)

            fitMol.SetTitle(fitTitle)
            oedepict.OEPrepareDepiction(fitMol)
            #
            nAtomsRef = refMol.NumAtoms()
            nAtomsFit = fitMol.NumAtoms()
            minAtoms = min(nAtomsRef, nAtomsFit)
            self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))

            # scaling
            refscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            fitscale = oedepict.OEGetMoleculeScale(fitMol, self._opts)
            self._opts.SetScale(min(refscale, fitscale))

            unique = True
            self._miter = self._mcss.Match(fitMol, unique)
            if self._miter.IsValid():
                match = self._miter.Target()
                oedepict.OEPrepareAlignedDepiction(fitMol, self._mcss.GetPattern(), match)

                # Depict reference molecule with MCS highlighting
                if (firstOne and layout in ["list"]) or layout in ["pairs"]:
                    firstOne = False
                    if layout in ["pairs"]:
                        refdisp = self._setHighlightStyleRef(matchObj=match, refMol=self._mcss.GetPattern())
                    else:
                        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
                    oedepict.OERenderMolecule(self.__imageRef, refdisp)
                    fU.mkdirForFile(refImagePath)
                    oedepict.OEWriteImage(refImagePath, self.__imageRef)

                # Depict fit molecule with MCS highlighting
                fitdisp = self._setHighlightStyleFit(matchObj=match, fitMol=fitMol)
                oedepict.OERenderMolecule(self.__imageFit, fitdisp)
                fU.mkdirForFile(fitImagePath)
                oedepict.OEWriteImage(fitImagePath, self.__imageFit)

                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        return atomMap


class OeMCSAlignUtil(OeDepictAlignBase):
    """Perform MCSS alignments.  Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.
    """

    def __init__(self, **kwargs):
        super(OeMCSAlignUtil, self).__init__(**kwargs)
        #
        self.__maxMatches = kwargs.get("maxMatches", 2048)

    def doAlign(self):
        """Test the MCSS comparison between the current reference and fit molecules -

        Return list of corresponding atoms on success or an empty list otherwise.
        """
        atomMap = []
        self._setupMCSS(self._refMol)
        #
        nAtomsRef = self._refMol.NumAtoms()
        nAtomsFit = self._fitMol.NumAtoms()
        minAtoms = min(nAtomsRef, nAtomsFit)
        self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))
        unique = True
        self._miter = self._mcss.Match(self._fitMol, unique)
        if self._miter.IsValid():
            match = self._miter.Target()
            for mAt in match.GetAtoms():
                atomMap.append((self._refId, mAt.pattern.GetName(), self._fitId, mAt.target.GetName()))

        return atomMap

    def __getAtomIndex(self, oeMol):
        atD = {}
        try:
            for ii, atom in enumerate(oeMol.GetAtoms()):
                atD[atom.GetName().strip()] = ii
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError):
            pass
        return atD

    def __getChargeIndex(self, oeMol):
        atD = {}
        try:
            for atom in oeMol.GetAtoms():
                atD[atom.GetName().strip()] = atom.GetFormalCharge()
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError):
            pass
        return atD

    def __getNeighbors(self, oeMol, atNameList):
        nL = []
        try:
            for atom in oeMol.GetAtoms():
                aN = atom.GetName().strip()
                if aN in atNameList:
                    tL = []
                    for nbr in atom.GetAtoms():
                        tL.append(nbr.GetName())
                        nL.append((aN, tL))
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError):
            pass
        return nL

    def doTestAlign(self):
        """Test the MCSS comparison between the current reference and fit molecules -

        Return list of corresponding atoms on success or an empty list otherwise.
        """
        self._minAtomMatchFraction = 0.9
        unique = True
        atomMap = []
        self._setupMCSS(self._refMol)
        #
        nAtomsRef = self._refMol.NumAtoms()
        nAtomsFit = self._fitMol.NumAtoms()
        #
        #
        minAtoms = min(nAtomsRef, nAtomsFit)
        self._mcss.SetMCSFunc(oechem.OEMCSMaxAtoms())
        self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))
        unique = False

        self._miter = self._mcss.Match(self._fitMol, unique)
        if self._miter.IsValid():
            match = self._miter.Target()
            for mAt in match.GetAtoms():
                atomMap.append((self._refId, mAt.pattern.GetName(), self._fitId, mAt.target.GetName()))

        logger.debug("nAtomsRef %d nAtomsFit %d match length %d", nAtomsRef, nAtomsFit, len(atomMap))

        return atomMap

    def doAlignWithAnal(self):
        """Test the MCSS comparison between the current reference and fit molecules -

        Return list of corresponding atoms on success or an empty list otherwise.
        """
        self._minAtomMatchFraction = 0.9
        unique = True
        atomMap = []
        self._setupMCSS(self._refMol)
        #
        nAtomsRef = self._refMol.NumAtoms()
        nAtomsFit = self._fitMol.NumAtoms()
        #
        atomRefD = self.__getAtomIndex(self._refMol)
        atomFitD = self.__getAtomIndex(self._fitMol)
        #
        chgRefD = self.__getChargeIndex(self._refMol)
        chgFitD = self.__getChargeIndex(self._fitMol)
        #
        minAtoms = min(nAtomsRef, nAtomsFit)
        #
        self._mcss.SetMaxMatches(self.__maxMatches)
        nlim = oechem.OEGetMCSExhaustiveSearchTruncationLimit()
        logger.debug("search limit %d max matches %d", nlim, self.__maxMatches)
        self._mcss.SetMCSFunc(oechem.OEMCSMaxAtoms())
        self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))
        unique = False

        atomRefMatchD = {}
        atomFitMatchD = {}
        self._miter = self._mcss.Match(self._fitMol, unique)
        if self._miter.IsValid():
            match = self._miter.Target()
            for mAt in match.GetAtoms():
                atomMap.append((self._refId, mAt.pattern.GetName(), self._fitId, mAt.target.GetName()))
                atomRefMatchD[mAt.pattern.GetName()] = mAt.target.GetName().strip()
                atomFitMatchD[mAt.target.GetName()] = mAt.pattern.GetName().strip()
        logger.debug("nAtomsRef %d nAtomsFit %d match length %d", nAtomsRef, nAtomsFit, len(atomMap))

        #
        # Get unmapped reference and fit atoms -
        #
        uRefAtomList = []
        uRefNList = []
        for atN in atomRefD:
            if atN not in atomRefMatchD:
                uRefAtomList.append(atN)
        #
        if len(uRefAtomList) > 0:
            uRefNList = self.__getNeighbors(self._refMol, uRefAtomList)
        #
        uFitAtomList = []
        uFitNList = []
        for atN in atomFitD:
            if atN not in atomFitMatchD:
                uFitAtomList.append(atN)
        #
        #
        if len(uFitAtomList) > 0:
            uFitNList = self.__getNeighbors(self._fitMol, uFitAtomList)
        #
        chgDifRefD = {}
        for atN0 in chgRefD:
            if atN0 in atomRefMatchD:
                atN = atomRefMatchD[atN0]
                # if ((atN not in chgFitD) or (chgRefD[atN0] != chgFitD[atN])):
                if chgRefD[atN0] != chgFitD[atN]:
                    chgDifRefD[atN] = chgRefD[atN0]
        #
        chgDifFitD = {}
        for atN0 in chgFitD:
            if atN0 in atomFitMatchD:
                atN = atomFitMatchD[atN0]
                # if ((atN not in chgRefD) or (chgFitD[atN0] != chgRefD[atN])):
                if chgFitD[atN0] != chgRefD[atN]:
                    chgDifFitD[atN] = chgFitD[atN0]

        return atomMap, uRefAtomList, uRefNList, chgDifRefD, uFitAtomList, uFitNList, chgDifFitD

    def doAlignList(self):
        """Test MCSS comparison between the current reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,fitId,fitMol,fitTitle)

        Map of corresponding atoms is returned.

        """
        atomMap = []
        for (refId, refMol, _, _, fitId, fitMol, _, _) in self._pairMolList:
            #
            self._setupMCSS(refMol)
            #
            #
            nAtomsRef = refMol.NumAtoms()
            nAtomsFit = fitMol.NumAtoms()
            minAtoms = min(nAtomsRef, nAtomsFit)
            self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))

            unique = True
            self._miter = self._mcss.Match(fitMol, unique)

            if self._miter.IsValid():
                match = self._miter.Target()
                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        return atomMap


class OeSubStructureAlignUtil(OeDepictAlignBase):
    """Perform substructure alignments.  Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.
    """

    def __init__(self, **kwargs):
        super(OeSubStructureAlignUtil, self).__init__(**kwargs)
        #
        self.__unique = kwargs.get("unique", True)
        self.__maxMatches = kwargs.get("maxMatches", 2048)

    def doAlign(self):
        """Test the MCSS comparison between the current reference and fit molecules -

        Return list of corresponding atoms on success or an empty list otherwise.
        """
        atomMap = []
        self._setupSubStructure(self._refMol)
        #
        unique = self.__unique
        self._ss.SetMaxMatches(self.__maxMatches)
        self._miter = self._ss.Match(self._fitMol, unique)
        if self._miter.IsValid():
            match = self._miter.Target()
            for mAt in match.GetAtoms():
                atomMap.append((self._refId, mAt.pattern.GetName(), self._fitId, mAt.target.GetName()))

        return atomMap

    def doAlignList(self):
        """Test MCSS comparison between the current reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,fitId,fitMol,fitTitle)

        Map of corresponding atoms is returned.

        """
        atomMap = []
        unique = self.__unique
        for (refId, refMol, _, _, fitId, fitMol, _, _) in self._pairMolList:
            #
            self._setupSubStructure(refMol)
            #
            self._ss.SetMaxMatches(self.__maxMatches)
            self._miter = self._ss.Match(fitMol, unique)
            if self._miter.IsValid():
                match = self._miter.Target()
                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        return atomMap


class OeDepictSubStructureAlign(OeDepictAlignBase):
    """Create 2D depictions of substructure alignments. Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.

    Outputs are separate image files with a single diagram per file.
    """

    def __init__(self, **kwargs):
        super(OeDepictSubStructureAlign, self).__init__(**kwargs)
        #
        self.__imageRef = None
        self.__imageFit = None
        #

    def __setupImage(self):
        """Internal method to configure a single page image."""
        #
        self.__imageRef = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        self.__imageFit = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        self._opts = oedepict.OE2DMolDisplayOptions(self.__imageRef.GetWidth(), self.__imageRef.GetHeight(), oedepict.OEScale_AutoScale)
        self._assignDisplayOptions()

    def alignPair(self, maxMatches=10):
        """Depict a single aligned ref/fit molecule pair or the first ref/fit molecule pair on the
        current _pairMolList.  Display options set for a single grid row with two columns.
        """
        aM = []
        self._pairMolList = []
        self._pairMolList.append((self._refId, self._refMol, self._refTitle, self._refImagePath, self._fitId, self._fitMol, self._fitTitle, self._fitImagePath))
        try:
            aM = self.__alignListWorker(layout="pairs", maxMatches=maxMatches)
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignPairList(self, maxMatches=10):
        aM = []
        try:
            aM = self.__alignListWorker(layout="pairs", maxMatches=maxMatches)
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignOneWithList(self, maxMatches=10):
        aM = []
        try:
            aM = self.__alignListWorker(layout="list", maxMatches=maxMatches)
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))
        return aM

    @timeout("instance.timeOut", use_signals=False, dec_allow_eval=True)
    def __alignListWorker(self, layout="pairs", maxMatches=10):
        """Working method comparing a reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,refImagePath, fitId,fitMol,fitTitle, fitImagePath)

        Map of corresponding atoms is returned.

        Writes separate output images for the reference and fit molecules for each comparison pair.
        """
        #
        atomMap = []
        firstOne = True
        fU = FileUtil()
        for (refId, refMol, _, refImagePath, fitId, fitMol, fitTitle, fitImagePath) in self._pairMolList:
            #
            self.__setupImage()
            #
            oedepict.OEPrepareDepiction(refMol)
            #
            self._setupSubStructure(refMol)
            fitMol.SetTitle(fitTitle)
            oedepict.OEPrepareDepiction(fitMol)
            #
            # nAtomsRef = refMol.NumAtoms()
            # nAtomsFit = fitMol.NumAtoms()
            # minAtoms = min(nAtomsRef, nAtomsFit)
            # self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))

            # scaling
            refscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            fitscale = oedepict.OEGetMoleculeScale(fitMol, self._opts)
            self._opts.SetScale(min(refscale, fitscale))
            #
            unique = True
            self._ss.SetMaxMatches(maxMatches)
            self._miter = self._ss.Match(fitMol, unique)
            if self._miter.IsValid():
                match = self._miter.Target()
                oedepict.OEPrepareAlignedDepiction(fitMol, self._ss.GetPattern(), match)

                # Depict reference molecule with SS highlighting
                if (firstOne and layout in ["list"]) or layout in ["pairs"]:
                    firstOne = False
                    if layout in ["pairs"]:
                        refdisp = self._setHighlightStyleRef(matchObj=match, refMol=self._ss.GetPattern())
                    else:
                        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
                    logger.info("writing %r", refImagePath)
                    oedepict.OERenderMolecule(self.__imageRef, refdisp)
                    fU.mkdirForFile(refImagePath)
                    oedepict.OEWriteImage(refImagePath, self.__imageRef)

                # Depict fit molecule with SS highlighting
                fitdisp = self._setHighlightStyleFit(matchObj=match, fitMol=fitMol)
                oedepict.OERenderMolecule(self.__imageFit, fitdisp)
                fU.mkdirForFile(fitImagePath)
                oedepict.OEWriteImage(fitImagePath, self.__imageFit)

                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        return atomMap


#  --
class OeDepictSubStructureAlignMultiPage(OeDepictAlignBase):
    """Create 2D depictions of substructure alignments. Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.

    Output images are rendered in a grid layout that can span multiple pages.
    """

    def __init__(self, **kwargs):
        super(OeDepictSubStructureAlignMultiPage, self).__init__(**kwargs)
        self.__grid = None
        self.__gridRows = None
        self.__gridCols = None
        self.__multi = None
        self.__image = None
        self.__citer = None
        #
        #

    def alignPairListMulti(self, imagePath="multi.pdf", maxMatches=10):
        aM = []
        self._params["gridCols"] = 2
        try:
            aM = self.__alignListMultiWorker(imagePath=imagePath, layout="pairs", maxMatches=maxMatches)
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))

        return aM

    def alignOneWithListMulti(self, imagePath="multi.pdf", maxMatches=10):
        aM = []
        try:
            aM = self.__alignListMultiWorker(imagePath=imagePath, layout="list", maxMatches=maxMatches)
        except (IndexError, ValueError, AttributeError, TypeError, RuntimeError) as e:
            logger.exception("Failing with %s", str(e))

        return aM

    def __setupImageMulti(self):
        """Internal method to configure a multipage image."""
        #
        self.__gridRows = self._params["gridRows"]
        self.__gridCols = self._params["gridCols"]
        #
        if self._params["pageOrientation"] == "landscape":
            self.__multi = oedepict.OEMultiPageImageFile(oedepict.OEPageOrientation_Landscape, oedepict.OEPageSize_US_Letter)
        else:
            self.__multi = oedepict.OEMultiPageImageFile(oedepict.OEPageOrientation_Portrait, oedepict.OEPageSize_US_Letter)

        self.__newPage()

    def __newPage(self):
        """Internal method to advance to a new page in a multipage configuration."""
        rows = self.__gridRows
        cols = self.__gridCols
        self.__image = self.__multi.NewPage()
        self.__grid = oedepict.OEImageGrid(self.__image, rows, cols)
        self.__grid.SetCellGap(self._params["cellGap"])
        self.__grid.SetMargins(self._params["cellMargin"])
        logger.debug("Num columns %d", self.__grid.NumCols())
        logger.debug("Num rows    %d", self.__grid.NumRows())
        self._opts = oedepict.OE2DMolDisplayOptions(self.__grid.GetCellWidth(), self.__grid.GetCellHeight(), oedepict.OEScale_AutoScale)
        self._assignDisplayOptions()
        self.__citer = self.__grid.GetCells()

    @timeout("instance.timeOut", use_signals=False, dec_allow_eval=True)
    def __alignListMultiWorker(self, imagePath="multi.pdf", layout="pairs", maxMatches=10):
        """Working method comparing a reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,refImagePath,fitId,fitMol,fitTitle,fitImagePath)

        Map of corresponding atoms is returned.

        Image Output is in multipage layout.
        """
        #
        self.__setupImageMulti()
        #
        atomMap = []
        firstOne = True
        iCount = 0
        for (refId, refMol, _, _, fitId, fitMol, fitTitle, _) in self._pairMolList:
            iCount += 1
            #
            oedepict.OEPrepareDepiction(refMol)
            self._setupSubStructure(refMol)
            #
            #
            fitMol.SetTitle(fitTitle)
            oedepict.OEPrepareDepiction(fitMol)
            #
            # nAtomsRef = refMol.NumAtoms()
            # nAtomsFit = fitMol.NumAtoms()
            # minAtoms = min(nAtomsRef, nAtomsFit)
            # mcssMinAtoms = int(minAtoms * self._minAtomMatchFraction)
            # self._mcss.SetMinAtoms(mcssMinAtoms)

            # scaling
            refscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            fitscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            self._opts.SetScale(min(refscale, fitscale))

            unique = True
            self._ss.SetMaxMatches(maxMatches)
            self._miter = self._ss.Match(fitMol, unique)
            logger.debug("ss match completed for refId %s fitId %s", refId, fitId)
            if self._miter.IsValid():
                match = self._miter.Target()
                oedepict.OEPrepareAlignedDepiction(fitMol, self._ss.GetPattern(), match)

                # Depict reference molecule with SS highlighting
                if (firstOne and layout in ["list"]) or layout in ["pairs"]:
                    firstOne = False
                    if layout in ["pairs"]:
                        refdisp = self._setHighlightStyleRef(matchObj=match, refMol=self._ss.GetPattern())
                    else:
                        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
                    if not self.__citer.IsValid():
                        self.__newPage()
                    cell = self.__citer.Target()
                    self.__citer.Next()
                    oedepict.OERenderMolecule(cell, refdisp)
                    if self._params["cellBorders"]:
                        oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                # Depict fit molecule with SS highlighting
                fitdisp = self._setHighlightStyleFit(matchObj=match, fitMol=fitMol)
                if not self.__citer.IsValid():
                    self.__newPage()
                cell = self.__citer.Target()
                self.__citer.Next()
                oedepict.OERenderMolecule(cell, fitdisp)

                if self._params["cellBorders"]:
                    oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

                logger.debug("mcss match completed for refId %s fitId %s total map length %d", refId, fitId, len(atomMap))

        logger.debug("writing image %s", imagePath)
        fU = FileUtil()
        fU.mkdirForFile(imagePath)
        oedepict.OEWriteMultiPageImage(imagePath, self.__multi)
        logger.debug("completed with map lenth %d", len(atomMap))
        #
        return atomMap



##
# File:    ChemCompMoleculeProviderTests.py
# Author:  J. Westbrook
# Date:    1-Oct-2019
# Version: 0.001
#
# Update:
#  5-Mar-2020 jdw all cases including filtered subsets working with naming prefixes.
#
##
"""
Tests for utilities to read and process the dictionary of PDB chemical component definitions.

"""
import logging
import os
import platform
import resource
import time
import unittest

from importlib.metadata import version as get_package_version
from rcsb.utils.chem.ChemCompMoleculeProvider import ChemCompMoleculeProvider
from rcsb.utils.io.MarshalUtil import MarshalUtil

__version__ = get_package_version("rcsb.utils.chem")

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class ChemCompMoleculeProviderTests(unittest.TestCase):
    skipFlag = True

    def setUp(self):
        self.__startTime = time.time()
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        self.__missedIdsPath = os.path.join(self.__dataPath, "filteredIds.json")
        #
        logger.debug("Running tests on version %s", __version__)
        logger.info("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testBuildMoleculeCacheFilesAbbrev(self):
        """Test construction of abbreviated of chemical component definitions from alternate source paths"""
        self.__testBuildMoleculeCacheFiles(ccUrlTarget=self.__ccUrlTarget, birdUrlTarget=self.__birdUrlTarget, ccFileNamePrefix="cc-abbrev", skipObsolete=True)

    @unittest.skipIf(skipFlag, "Long test")
    def testBuildMoleculeCacheFilesSubset(self):
        """Test construction of a subset (1K) of chemical component definitions."""
        self.__testBuildMoleculeCacheFiles(molLimit=1000, ccFileNamePrefix="cc-1k")
        #

    @unittest.skipIf(skipFlag, "Long test")
    def testSubsetBuildMoleculeCacheFiltered(self):
        """Test construction of a filtered selection of chemical component definitions."""
        mU = MarshalUtil()
        fD = mU.doImport(self.__missedIdsPath, fmt="json")
        filterIdD = {ccId: True for ccId in fD["filteredIdList"]}
        self.__testBuildMoleculeCacheFiles(filterIdD=filterIdD, ccFileNamePrefix="cc-filtered")
        #

    @unittest.skipIf(skipFlag, "Long test")
    def testBuildMoleculeCacheFilesFull(self):
        """Test construction and reload of full chemical component resource files.
        Running 151 sec sp macbookpro 6.6G resident mem
        """
        self.__testBuildMoleculeCacheFiles(logSizes=True)
        logger.info("Reloading full cache file")
        self.__testBuildMoleculeCacheFiles(useCache=True, logSizes=False)

    # ----
    def __testBuildMoleculeCacheFiles(self, **kwargs):
        """Test build chemical component cache files from the input component dictionaries"""
        try:
            ccUrlTarget = kwargs.get("ccUrlTarget", None)
            birdUrlTarget = kwargs.get("birdUrlTarget", None)
            molLimit = kwargs.get("molLimit", None)
            minCount = kwargs.get("minCount", None)
            useCache = kwargs.get("useCache", False)
            logSizes = kwargs.get("logSizes", False)
            filterIdD = kwargs.get("filterIdD", None)
            skipObsolete = kwargs.get("skipObsolete", True)
            ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "cc")
            #
            if ccUrlTarget and birdUrlTarget:
                ccmP = ChemCompMoleculeProvider(
                    ccUrlTarget=ccUrlTarget,
                    birdUrlTarget=birdUrlTarget,
                    ccFileNamePrefix=ccFileNamePrefix,
                    cachePath=self.__cachePath,
                    useCache=useCache,
                    molLimit=molLimit,
                    filterIdD=filterIdD,
                    skipObsolete=skipObsolete,
                )
                ok = ccmP.testCache(minCount=molLimit, logSizes=logSizes)
                self.assertTrue(ok)
            else:
                ccmP = ChemCompMoleculeProvider(cachePath=self.__cachePath, useCache=useCache, ccFileNamePrefix=ccFileNamePrefix, molLimit=molLimit, filterIdD=filterIdD)
                ok = ccmP.testCache(minCount=minCount, logSizes=logSizes)
                self.assertTrue(ok)
        except Exception as e:
            logger.info("Failing with %s", str(e))
            self.fail()


def buildCacheFiles():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(ChemCompMoleculeProviderTests("testBuildMoleculeCacheFilesAbbrev"))
    suiteSelect.addTest(ChemCompMoleculeProviderTests("testBuildMoleculeCacheFilesSubset"))
    suiteSelect.addTest(ChemCompMoleculeProviderTests("testBuildMoleculeCacheFilesFiltered"))
    suiteSelect.addTest(ChemCompMoleculeProviderTests("testBuildMoleculeCacheFilesFull"))
    return suiteSelect


if __name__ == "__main__":

    mySuite = buildCacheFiles()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    OeSubStructSearchCompareTests.py
# Author:  J. Westbrook
# Date:    26-Feb-2021
# Version: 0.001
#
# Update:
#
#
##
"""
Match and Substructure search comparison tests on the full data including display.
"""
import logging
import os
import time
import unittest

from importlib.metadata import version as get_package_version
from rcsb.utils.chem.ChemCompMoleculeProvider import ChemCompMoleculeProvider
from rcsb.utils.chem.ChemCompSearchIndexProvider import ChemCompSearchIndexProvider
from rcsb.utils.chem.OeDepictAlign import OeDepictMCSAlignPage, OeDepictMCSAlignMultiPage, OeDepictSubStructureAlignMultiPage
from rcsb.utils.chem.OeIoUtils import OeIoUtils
from rcsb.utils.chem.OeSearchMoleculeProvider import OeSearchMoleculeProvider
from rcsb.utils.chem.OeSearchUtils import OeSearchUtils
from rcsb.utils.chem.OeSubStructSearchUtils import OeSubStructSearchUtils

__version__ = get_package_version("rcsb.utils.chem")

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()
logger.setLevel(logging.INFO)


class OeSubStructSearchCompareTests(unittest.TestCase):
    useFull = False

    def setUp(self):

        self.__workPath = os.path.join(HERE, "test-output")
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        self.__doDisplay = True
        self.__numProcPrep = 6
        self.__numProcSearch = 6
        self.__minCount = None
        self.__startTime = time.time()
        #
        if OeSubStructSearchCompareTests.useFull:
            self.__myKwargs = {
                "cachePath": self.__cachePath,
                "useCache": True,
                "ccFileNamePrefix": "cc-full",
                "oeFileNamePrefix": "oe-full",
                "molBuildType": "model-xyz",
                "limitPerceptions": False,
                "screenTypeList": None,
                "numProc": self.__numProcPrep,
                "suppressHydrogens": True,
                "matchOpts": "sub-struct-graph-relaxed",
                "fpTypeCuttoffD": {"TREE": 0.6, "MACCS": 0.9},
                "maxFpResults": 50,
            }
        else:
            self.__myKwargs = {
                "ccUrlTarget": self.__ccUrlTarget,
                "birdUrlTarget": self.__birdUrlTarget,
                "cachePath": self.__cachePath,
                "useCache": True,
                "ccFileNamePrefix": "cc-abbrev",
                "oeFileNamePrefix": "oe-abbrev",
                "molBuildType": "model-xyz",
                "limitPerceptions": False,
                "screenTypeList": None,
                "numProc": self.__numProcPrep,
                "suppressHydrogens": True,
                "matchOpts": "sub-struct-graph-relaxed",
                "fpTypeCuttoffD": {"TREE": 0.6, "MACCS": 0.9},
                "maxFpResults": 50,
            }
        #
        self.__oesmP = OeSearchMoleculeProvider(**self.__myKwargs)
        ok = self.__oesmP.testCache()
        self.assertTrue(ok)
        #
        self.__ccmP = ChemCompMoleculeProvider(**self.__myKwargs)
        self.__ccmP.testCache()
        #
        self.__ccsidxP = ChemCompSearchIndexProvider(**self.__myKwargs)
        ok = self.__ccsidxP.testCache(minCount=self.__minCount)
        self.assertTrue(ok)
        self.__oessU = OeSubStructSearchUtils(self.__oesmP)
        ok = self.__oessU.testCache()
        self.assertTrue(ok)
        #
        fpTypeCuttoffD = self.__myKwargs.get("fpTypeCuttoffD", {})
        fpTypeList = [k for k, v in fpTypeCuttoffD.items()]
        self.__oesU = OeSearchUtils(self.__oesmP, fpTypeList=fpTypeList)
        ok = self.__oesU.testCache()
        self.assertTrue(ok)
        #
        logger.debug("Running tests on version %s", __version__)
        logger.info("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    @unittest.skipIf(not useFull, "Requires full data set")
    def testSubStructSearchDescriptor(self):
        #
        query = "n1ccccc1"
        queryId = "query-smiles"
        queryType = "oe-iso-smiles"
        #
        limitPerceptions = self.__myKwargs.get("limitPerceptions", False)
        suppressHydrogens = self.__myKwargs.get("suppressHydrogens", True)
        numProc = self.__myKwargs.get("numProc", 4)
        # for matchOpts in ["sub-struct-graph-relaxed", "sub-struct-graph-relaxed-stereo", "sub-struct-graph-strict"]:
        for matchOpts in ["sub-struct-graph-strict"]:
            #
            oeMol = self.__getMol(query, queryType, queryId, limitPerceptions=limitPerceptions, suppressHydrogens=suppressHydrogens)
            startTime = time.time()
            retStatus, mL = self.__search(oeMol, matchOpts, numProc)
            logger.info("%s status (%r) matchOpts %s result %d in (%.4f seconds)", queryId, retStatus, matchOpts, len(mL), time.time() - startTime)
            self.assertTrue(retStatus)
            if queryType == "CC":
                self.assertTrue(self.__resultContains(queryId, mL))
            #
            if self.__doDisplay:
                self.__display(mL, query, queryId, queryType, matchOpts)

    @unittest.skipIf(not useFull, "Requires full data set")
    def testSubStructSearchSelected(self):
        #
        query = queryId = "STI"
        queryType = "CC"
        #
        limitPerceptions = self.__myKwargs.get("limitPerceptions", False)
        suppressHydrogens = self.__myKwargs.get("suppressHydrogens", True)
        numProc = self.__myKwargs.get("numProc", 4)
        for matchOpts in ["sub-struct-graph-relaxed", "sub-struct-graph-relaxed-stereo", "sub-struct-graph-strict"]:
            #
            oeMol = self.__getMol(query, queryType, queryId, limitPerceptions=limitPerceptions, suppressHydrogens=suppressHydrogens)
            startTime = time.time()
            retStatus, mL = self.__search(oeMol, matchOpts, numProc)
            logger.info("%s status (%r) matchOpts %s result %d in (%.4f seconds)", queryId, retStatus, matchOpts, len(mL), time.time() - startTime)
            self.assertTrue(retStatus)
            if queryType == "CC":
                self.assertTrue(self.__resultContains(queryId, mL))
            #
            if self.__doDisplay:
                self.__display(mL, query, queryId, queryType, matchOpts)

    @unittest.skipIf(not useFull, "Requires full data set")
    def testSubStructSearchAll(self):
        #
        ccD = self.__ccmP.getMolD()
        for ccId in ccD:
            query = queryId = ccId
            if ccId in ["UNX", "UNL", "UNK", "DUM"]:
                continue
            queryType = "CC"
            #
            limitPerceptions = self.__myKwargs.get("limitPerceptions", False)
            suppressHydrogens = self.__myKwargs.get("suppressHydrogens", True)
            numProc = self.__myKwargs.get("numProc", 5)
            for matchOpts in ["sub-struct-graph-relaxed", "sub-struct-graph-relaxed-stereo", "sub-struct-graph-strict"]:
                #
                oeMol = self.__getMol(query, queryType, queryId, limitPerceptions=limitPerceptions, suppressHydrogens=suppressHydrogens)
                if oeMol.NumAtoms() < 3:
                    continue
                #
                startTime = time.time()
                retStatus, mL = self.__search(oeMol, matchOpts, numProc)
                logger.info("%s status (%r) matchOpts %s result %d in (%.4f seconds)", queryId, retStatus, matchOpts, len(mL), time.time() - startTime)
                self.assertTrue(retStatus)
                if queryType == "CC":
                    self.assertTrue(self.__resultContains(queryId, mL))
                #
                if self.__doDisplay:
                    self.__display(mL, query, queryId, queryType, matchOpts)

    @unittest.skipIf(not useFull, "Requires full data set")
    def testMatchSearchSelected(self):
        #
        query = queryId = "STI"
        queryType = "CC"
        #
        limitPerceptions = self.__myKwargs.get("limitPerceptions", False)
        suppressHydrogens = self.__myKwargs.get("suppressHydrogens", True)
        numProc = self.__myKwargs.get("numProc", 4)
        for matchOpts in ["fingerprint-similarity", "graph-relaxed", "graph-relaxed-stereo", "graph-strict"]:
            #
            oeMol = self.__getMol(query, queryType, queryId, limitPerceptions=limitPerceptions, suppressHydrogens=suppressHydrogens)
            startTime = time.time()
            retStatus, mL = self.__search(oeMol, matchOpts, numProc)
            logger.info("%s status (%r) matchOpts %s result %d in (%.4f seconds)", queryId, retStatus, matchOpts, len(mL), time.time() - startTime)
            self.assertTrue(retStatus)
            if queryType == "CC":
                self.assertTrue(self.__resultContains(queryId, mL))
            #
            if self.__doDisplay:
                self.__display(mL, query, queryId, queryType, matchOpts)

    def __search(self, oeMol, matchOpts, numProc):
        fpL = mL = None
        if matchOpts.startswith("sub-struct-"):
            retStatus, mL = self.__subStructureSearch(oeMol, matchOpts=matchOpts, numProc=numProc)
        else:
            retStatus, mL, fpL = self.__matchSearch(oeMol, matchOpts=matchOpts)
        #
        rL = fpL if matchOpts in ["fingerprint-similarity"] else mL
        return retStatus, rL

    #
    def __subStructureSearch(self, oeMol, matchOpts, numProc):
        ##
        ccIdL = self.__oessU.prefilterIndex(oeMol, self.__ccsidxP, matchOpts=matchOpts, skipFeatures=False)
        retStatus, mL = self.__oessU.searchSubStructure(oeMol, ccIdList=ccIdL, matchOpts=matchOpts, numProc=numProc)
        return retStatus, mL

    def __matchSearch(self, oeMol, matchOpts="graph-relaxed"):
        ssL = fpL = []
        try:
            fpTypeCuttoffD = self.__myKwargs.get("fpTypeCuttoffD", {})
            maxFpResults = self.__myKwargs.get("maxFpResults", 50)
            retStatus, ssL, fpL = self.__oesU.searchSubStructureAndFingerPrint(oeMol, list(fpTypeCuttoffD.items())[:2], maxFpResults, matchOpts=matchOpts)
            # logger.info("fpL %r", fpL)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            #
        return retStatus, ssL, fpL

    def __getMol(self, query, queryType, queryId, limitPerceptions=False, suppressHydrogens=True):
        oeioU = OeIoUtils()
        if queryType == "CC":
            oeMol = self.__oesmP.getMol(query)
        else:
            oeMol = oeioU.descriptorToMol(query, queryType, limitPerceptions=limitPerceptions, messageTag=queryId)
        #
        if suppressHydrogens:
            oeMol = oeioU.suppressHydrogens(oeMol)
        oeMol.SetTitle(queryId)
        return oeMol

    def __resultContains(self, ccId, matchResultList):
        for matchResult in matchResultList:
            if matchResult.ccId == ccId:
                return True
        return False

    #
    #  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------
    def __display(self, mL, query, queryId, queryType, matchOpts):
        smL = sorted(mL, key=lambda kv: kv.fpScore, reverse=True)
        # ----
        tD = {}
        for sm in smL:
            ccId = sm.ccId.split("|")[0]
            tD.setdefault(ccId, []).append(sm)
        dL = []
        for ccId, ttL in tD.items():
            if len(ttL) == 1:
                dL.append(ttL[0])
            else:
                parent = False
                for tt in ttL:
                    if tt.ccId == ccId:
                        dL.append(tt)
                        parent = True
                        break
                if not parent:
                    dL.append(ttL[0])
        # ----
        pdfImagePath = os.path.join(self.__workPath, queryId + "-" + matchOpts + ".pdf")
        self.__displayPaginatedAlignments(pdfImagePath, query, queryType, queryId, dL, matchOpts=matchOpts)

    def __displayPaginatedAlignments(self, pdfImagePath, query, queryType, queryId, matchResultList, matchOpts="relaxed-stereo", alignMode="SS"):
        refId = queryId
        oeMolRef = self.__getMol(query, queryType, queryId, limitPerceptions=False, suppressHydrogens=True)
        pairList = []
        for mr in sorted(matchResultList, key=lambda kv: kv.fpScore, reverse=True):
            fitId = mr.ccId.split("|")[0]
            if len(mr.ccId) > 4:
                fitId = fitId + " (tautomer/protomer)"
            oeMolFit = self.__oesmP.getMol(mr.ccId)
            pairList.append((refId, oeMolRef, fitId, oeMolFit))
        #
        self.__depictFitList(pdfImagePath, pairList, matchOpts=matchOpts, alignMode=alignMode)

    def __pairDepictPage(self, imagePath, refId, refTitle, refMol, fitId, fitTitle, fitMol, matchOpts="strict"):
        """Depict pairwise alignment of the input reference and fit molecules.

        Args:
            imagePath (str): path to image (format by path extension)
            refId (str): reference molecule identifier
            refTitle (str): reference molecule title
            refMol (obj): reference OE molecule object
            fitId (str): fit molecule identifier
            fitTitle (str): fit molecule title
            fitMol (obj): fit OE molecule object
            matchOpts (str, optional): alignment criteria (relaxed|relaxed-stereo|strict). Defaults to "strict".

        Returns:
            (list): atom mapping in all aligned figures
                    [(reference component Id, reference atom name, fit chemical component Id, fit atom name)
        """
        aML = []
        try:
            oed = OeDepictMCSAlignPage()
            oed.setSearchType(sType=matchOpts)

            oed.setRefMol(refMol, refId, title=refTitle)
            oed.setFitMol(fitMol, fitId, title=fitTitle)
            oed.setDisplayOptions(
                imageSizeX=2000,
                imageSizeY=1000,
                labelAtomName=True,
                labelAtomCIPStereo=True,
                labelAtomIndex=False,
                labelBondIndex=False,
                highlightStyleFit="ballAndStickInverse",
                bondDisplayWidth=0.5,
                highLightMatchColorRef="green",
                highLightNotMatchColorRef="pink",
            )
            aML = oed.alignPair(imagePath=imagePath)
            if aML:
                for (rCC, rAt, tCC, tAt) in aML:
                    logger.debug("%5s %-5s %5s %-5s", rCC, rAt, tCC, tAt)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aML

    def __depictFitList(self, pdfImagePath, pairList, matchOpts="exact", alignMode="SS"):
        """Depict pairwise alignments with multi-page layout in PDF format.

        Args:
            pdfImagePath (str): PDF image path
            pairList (list): [(refId, refOeMol, fitId, fitOeMol)]

        Returns:
            (list): atom mapping in all aligned figures
                    [(reference component Id, reference atom name, fit chemical component Id, fit atom name)
        """
        aML = []
        try:
            if alignMode == "MCSS":
                oed = OeDepictMCSAlignMultiPage()
            else:
                oed = OeDepictSubStructureAlignMultiPage()
            oed.setSearchType(sType=matchOpts)
            oed.setPairMolList(pairList)

            oed.setDisplayOptions(
                labelAtomName=True,
                labelAtomCIPStereo=True,
                labelAtomIndex=False,
                labelBondIndex=False,
                highlightStyleFit="ballAndStickInverse",
                pageOrientation="portrait",
                gridRows=4,
                bondDisplayWidth=0.5,
                highLightMatchColorRef="green",
                highLightNotMatchColorRef="pink",
            )
            aML = oed.alignPairListMulti(imagePath=pdfImagePath)
            if aML:
                for (rCC, rAt, tCC, tAt) in aML:
                    logger.debug("%5s %-5s %5s %-5s", rCC, rAt, tCC, tAt)
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aML


def searchComparison():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(OeSubStructSearchCompareTests("testSubStructureSearchFromIndex"))
    suiteSelect.addTest(OeSubStructSearchCompareTests("testSubStructureSearchBase"))
    suiteSelect.addTest(OeSubStructSearchCompareTests("testSubStructureSearchFromIndexBase"))
    suiteSelect.addTest(OeSubStructSearchCompareTests("testSubStructureSearchFromIndexSelected"))
    return suiteSelect


if __name__ == "__main__":
    mySuite = searchComparison()
    unittest.TextTestRunner(verbosity=2).run(mySuite)
